from typing import Any, Callable, List, Optional, Type, Union

import torch
import torch.nn as nn

from esp_ppq import PPQuantFunction, SingletonMeta
from esp_ppq.core import OperationQuantizationConfig, TargetPlatform, TensorQuantizationConfig, ppq_warning
from esp_ppq.IR import Operation, Variable
from esp_ppq.lib import FloatingQuantizationConfig, LinearQuantizationConfig, ParameterQuant, TensorQuant
from esp_ppq.quantization.quantizer import BaseQuantizer


class QuantLayer:
    def __init__(self) -> None:
        self._quantize_controler = None


class QATController:
    def __init__(self) -> None:
        self.export_mode = 'Plain Onnx'


class QConv1d(torch.nn.Conv1d, QuantLayer):
    def __init__(self, controller: QATController, **kwargs) -> None:
        super(torch.nn.Conv1d).__init__(**kwargs)
        self._quantize_controler = controller

        self.input_quant = TensorQuant(
            quant_config=LinearQuantizationConfig(
                symmetrical=True,
                power_of_2=False,
                channel_axis=None,  # channel_axis = None -> Per tensor Quantization
                calibration='percentile',
            )
        )

        self.weight_quant = ParameterQuant(
            quant_config=LinearQuantizationConfig(
                symmetrical=True,
                power_of_2=False,
                channel_axis=0,  # channel_axis = 1 -> Per channel Quantization
                calibration='minmax',
            )
        )

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        x = self.input_quant(input)
        w = self.weight_quant(self.weight)
        return self._conv_forward(x, w, self.bias)


class QConv2d(torch.nn.Conv2d, QuantLayer):
    def __init__(self, controller: QATController, **kwargs) -> None:
        super(torch.nn.Conv2d).__init__(**kwargs)
        self._quantize_controler = controller

        self.input_quant = TensorQuant(
            quant_config=LinearQuantizationConfig(
                symmetrical=True,
                power_of_2=False,
                channel_axis=None,  # channel_axis = None -> Per tensor Quantization
                calibration='percentile',
            )
        )

        self.weight_quant = ParameterQuant(
            quant_config=LinearQuantizationConfig(
                symmetrical=True,
                power_of_2=False,
                channel_axis=0,  # channel_axis = 1 -> Per channel Quantization
                calibration='minmax',
            )
        )

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        x = self.input_quant(input)
        w = self.weight_quant(self.weight)
        return self._conv_forward(x, w, self.bias)


class QConv3d(torch.nn.Conv3d, QuantLayer):
    def __init__(self, controller: QATController, **kwargs) -> None:
        super(torch.nn.Conv3d).__init__(**kwargs)
        self._quantize_controler = controller

        self.input_quant = TensorQuant(
            quant_config=LinearQuantizationConfig(
                symmetrical=True,
                power_of_2=False,
                channel_axis=None,  # channel_axis = None -> Per tensor Quantization
                calibration='percentile',
            )
        )

        self.weight_quant = ParameterQuant(
            quant_config=LinearQuantizationConfig(
                symmetrical=True,
                power_of_2=False,
                channel_axis=1,  # channel_axis = 1 -> Per channel Quantization
                calibration='minmax',
            )
        )

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        x = self.input_quant(input)
        w = self.weight_quant(self.weight)
        return self._conv_forward(x, w, self.bias)


class ENABLE_CALIBRATION:
    """ """

    def __init__(self, controler: QATController) -> None:
        pass

    def __enter__(self):
        pass

    def __exit__(self, *args):
        pass



import random
from random import randint
from typing import Iterable, List, Tuple

import torch
from torch.autograd import Function

from esp_ppq.core import (
    NUM_OF_CHECKPOINT_FETCHS,
    PPQ_CONFIG,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TensorQuantizationConfig,
)
from esp_ppq.executor import TorchQuantizeDelegator
from esp_ppq.IR import BaseGraph, Operation, SearchableGraph, Variable
from esp_ppq.quantization.qfunction import PPQuantFunction
from esp_ppq.utils.fetch import batch_random_fetch
from esp_ppq.utils.round import ppq_tensor_round


class CuLSQ_LT(Function):
    @staticmethod
    def forward(
        ctx,
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        quant_min: int,
        quant_max: int,
        rounding: RoundingPolicy,
    ) -> torch.Tensor:
        if not PPQ_CONFIG.USING_CUDA_KERNEL:
            raise PermissionError('Can not invoke CuLSQ, Cuda kernel is not compiled.')
        from esp_ppq.core import CUDA

        # quantization function, pure cuda implmentation
        quantized = CUDA.LinearQuantize_T(
            tensor=tensor, scales=scales, offsets=offsets, minimum=quant_min, maximum=quant_max, rounding=rounding.value
        )
        # https://pytorch.org/docs/stable/generated/torch.autograd.function.FunctionCtx.save_for_backward.html
        ctx.save_for_backward(tensor, scales, offsets)
        ctx._quant_params = [quant_min, quant_max, rounding.value]
        return quantized

    @staticmethod
    def backward(ctx, dy: torch.Tensor):
        if not PPQ_CONFIG.USING_CUDA_KERNEL:
            raise PermissionError('Can not invoke CuLSQ, Cuda kernel is not compiled.')
        from esp_ppq.core import CUDA

        dy = dy.contiguous()
        tensor, scales, offsets = ctx.saved_tensors
        quant_min, quant_max, rounding = ctx._quant_params
        dx, ds = CUDA.LinearQuantize_T_B(tensor, scales, offsets, dy, quant_min, quant_max, rounding)
        return dx, ds, None, None, None, None, None


class CuLSQ_LC(Function):
    @staticmethod
    def forward(
        ctx,
        tensor: torch.Tensor,
        scales: torch.Tensor,
        offsets: torch.Tensor,
        channel_axis: int,
        quant_min: int,
        quant_max: int,
        rounding: RoundingPolicy,
    ) -> torch.Tensor:
        if not PPQ_CONFIG.USING_CUDA_KERNEL:
            raise PermissionError('Can not invoke CuLSQ, PPQ_CONFIG.USING_CUDA_KERNEL = False.')
        from esp_ppq.core import CUDA

        quantized = CUDA.LinearQuantize_C(
            tensor=tensor,
            scales=scales,
            offsets=offsets,
            channel_axis=channel_axis,
            minimum=quant_min,
            maximum=quant_max,
            rounding=rounding.value,
        )
        # https://pytorch.org/docs/stable/generated/torch.autograd.function.FunctionCtx.save_for_backward.html
        ctx.save_for_backward(tensor, scales, offsets)
        ctx._quant_params = [quant_min, quant_max, channel_axis, rounding.value]
        return quantized

    @staticmethod
    def backward(ctx, dy: torch.Tensor):
        if not PPQ_CONFIG.USING_CUDA_KERNEL:
            raise PermissionError('Can not invoke CuLSQ, PPQ_CONFIG.USING_CUDA_KERNEL = False.')
        from esp_ppq.core import CUDA

        dy = dy.contiguous()
        tensor, scales, offsets = ctx.saved_tensors
        quant_min, quant_max, channel_axis, rounding = ctx._quant_params
        dx, ds = CUDA.LinearQuantize_C_B(tensor, scales, offsets, dy, quant_min, quant_max, channel_axis, rounding)
        return dx, ds, None, None, None, None, None, None


class FinetuneCheckPoint:
    """Finetune Check Point stores training loss for variables. It bounds to a
    specific variable, collects and stores its fp32 value as a reference.

    ATTENTION: collecting fp32 value might cause GPU memory overflow, so we use a seed to
        sample only a part of fp32 value instead(randomly pick about 2000 values from given tensor).

    Finetune Check Point maintains a seed for data collecting, a best loss, and a reference values.
    """

    def __init__(
        self, variable: str, random_fetch: bool = True, seed: int = None, fetchs: int = NUM_OF_CHECKPOINT_FETCHS
    ) -> None:
        if seed is None:
            seed = randint(0, 0xFFFFFFFF)
        self.monitor_var = variable
        self.best_loss = float(1e9)
        self.seed = seed
        self.references = []
        self.outputs = []
        self.fetchs = fetchs
        self.random_fetch = random_fetch

    def push(self, tensor: torch.Tensor, is_reference: bool) -> None:
        if self.random_fetch:
            tensor = batch_random_fetch(tensor, seed=self.seed, fetches_per_batch=self.fetchs)
        if is_reference:
            self.references.append(tensor)
        else:
            self.outputs.append(tensor)

    def pop(self) -> Tuple[torch.Tensor]:
        assert len(self.outputs) == len(self.references), (
            'Inconsistent samples detected.'
            f'Reference output gets {len(self.references)} samples, however output has {len(self.outputs)}.'
        )

        return self.outputs, self.references

    def clear(self):
        self.outputs.clear()


class RandomMemDataset:
    """A very little helper class for randomly pick data samples from your
    dataset."""

    def __init__(self, data: Iterable) -> None:
        self._data = data
        self._num_of_batchs = len(data)

    def pop(self):
        idx = random.randint(0, self._num_of_batchs - 1)
        return self._data[idx]


class PriorityQueue:
    """一个很低端的优先队列实现.

    因为python自带的那个实现少一个我需要的接口

    所以我就自己写了这个，它很慢，但是够用。
    """

    def __init__(
        self,
    ) -> None:
        self._data = []
        self._ops = set()
        self._idx = 0
        self._lazy_tag = True  # 延迟操作标志

    def pop(self) -> Tuple[int, Operation]:
        if not self._lazy_tag:
            self._data = sorted(self._data, key=lambda x: x[0])
            self._lazy_tag = True
        if self._idx >= len(self._data):
            raise IndexError('Index out of range!')
        ele = self._data[self._idx]
        self._idx += 1
        return ele

    def push(self, depth: int, op: Operation):
        if op in self._ops:
            return
        self._data.append((depth, op))
        self._ops.add(op)
        self._lazy_tag = False

    def empty(self) -> bool:
        return self._idx >= len(self._data)


class TrainableBlock:
    """TrainableBlock refers to a limited subgraph extracted from integrated
    computational graph. TrainableBlock have exact one input node and one
    output node, while its depth(the distance from input node to output node)
    is limited.

    Formal definition of TrainableBlock can be found with following code of BlockBuilder

    Minimal TrainableBlock is {p, p, {p}}, this block have only one node as both input and output.
    """

    def __init__(self, sp: Operation, ep: Operation, rps: List[Operation]) -> None:
        self.sp = sp  # 起始节点
        self.ep = ep  # 终止节点
        self.rps = rps  # 中继节点

    def __str__(self) -> str:
        return f'[Graph Block from {self.sp.name} to {self.ep.name}]'


class BlockBuilder:
    """
    Network Block Builder will cut your graph into small blocks(subgraphs)
    Each block will have exact 1 input operation and 1 output operation.

    Besides, there is a parameter 'limit' to control the size of each block.
    """

    def __init__(self, graph: BaseGraph, topo_order: List[Operation]) -> None:
        self.graph = graph
        self.op_orders = topo_order
        self.depth = {}
        self.search_engine = SearchableGraph(self.graph)
        self.initialize_depth()

    def create_block(self, sp: Operation, ep: Operation) -> TrainableBlock:
        if sp == ep:
            return TrainableBlock(sp=sp, ep=ep, rps=[sp])
        rps = self.search_engine.opset_matching(
            sp_expr=lambda x: x == sp, rp_expr=lambda x, y: True, ep_expr=lambda x: x == ep, direction='down'
        )
        rps = [(self.op_orders.index(op), op) for op in rps]
        rps = sorted(rps)
        return TrainableBlock(sp=sp, ep=ep, rps=[op for _, op in rps])

    def build(self, op: Operation, limit: int) -> TrainableBlock:
        """子图分割算法, 这个算法将从指定节点出发, 构造一个满足定义的子图结构 Solving best block from given
        operation.

        Block definition:(子图定义)
            A Block is a triple contains S, E, M,
                where S is the input node of block
                where E is the output node of block
                where M contains all nodes inside block

        Property:(子图性质)
            1. Minimal TrainableBlock start from p is {p, p, {p}},
               this block have only one node as both input and output.
            2. When S != E,
               E must on every path from S to graph output.
               S must on every path from graph input to E.
            3. M contains and only contains nodes on all paths from S to E.

        Lemma:(算法引理)
            1. 如果 s 的后继节点只有一个 e, 且 e 的输入只有一个，那么 {s, e, {s, e}} 构成满足定义的子图，从 s 寻找子图的任务可以递归由 e 完成
            2. 如果 s 的后继存在多个节点，则只存在两种情况:
                2.1 从 s 出发，最大子图即为 {s, s, {s}}。
                2.2 从 s 出发，可以构成子图 {s, e, R} (e!=s), 那么R中必须含有一个节点接收多个输入。（可用反证法证明，略）

        Algorithm:(算法)
            Build(s, d):
                如果区块长度大于所需，则返回现有内容
                从 s 出发，如果 s 的后继节点只有一个e，则判断e的输入节点个数：
                    1. 如果 e 是单输入节点，执行Build(e, d-1)，并将其结果与 {s, e, {s, e}} 合并
                    2. 如果 e 是多输入节点，算法立即停机，返回 {s, s, {s}}

                如果 s 的后继节点存在多个，找出距离 s 拓扑序最近的多输入的节点 k1，判断 s 到输出的路径是否能够被 k1 阻断
                    如果 k 成功阻断所有输出，执行Build(k1, d-1)，并将其结果与 {s, k1, F(s, k1)} 合并
                    如果 k 不能阻断输出，寻找距离 s 次近的多输入节点 k2，重复判断
                直到 kn 到达 s 的距离超出限制

                函数 F(s, k1) 取出 从 s 到 k1 路上的所有节点

        可利用引理证明算法正确性，从略
        时间复杂度: O(kd) k 为节点最大度数 d 为深度限制。建立所有Block所需时间 O(nkd)
        """

        def _find_multi_input_ep(op: Operation):
            # 如果当前节点后继节点存在多个，层序遍历寻找阻断节点
            least_first_queue = PriorityQueue()
            least_first_queue.push(self.depth[op], op)
            least_first_queue.pop()

            for down_op in self.graph.get_downstream_operations(op):
                least_first_queue.push(self.depth[down_op], down_op)

            while not least_first_queue.empty():
                iter_operation = least_first_queue.pop()[-1]
                if least_first_queue.empty():
                    upstream_ops = self.graph.get_upstream_operations(iter_operation)
                    if all([op in least_first_queue._ops for op in upstream_ops]) and len(upstream_ops) > 1:
                        return iter_operation
                for down_op in self.graph.get_downstream_operations(iter_operation):
                    least_first_queue.push(self.depth[down_op], down_op)

            # if least_first_queue is empty, it means we can not find an blocking ep from given sp.
            return None

        def _find_coherent_ep(op: Operation):
            # 如果当前节点后继节点只有一个，向下寻找直系节点
            # 但如果直系后继节点有多于一个输入，算法立即停机
            ops = self.graph.get_downstream_operations(op)
            if len(ops) == 1:
                following_op = ops[0]
                # PATCH 20220811，get_upstream_operations 不足以判断算子是否只有一个输入
                # 因为算子可以直接与图的 input 相连...
                non_parameter_input = following_op.num_of_input - following_op.num_of_parameter
                upstream_ops = len(self.graph.get_upstream_operations(following_op))
                if non_parameter_input == 1 and upstream_ops == 1:
                    return ops[0]
            return None

        sp, ep, future_ep = op, op, op
        while future_ep is not None:
            if len(self.graph.get_downstream_operations(ep)) <= 1:
                future_ep = _find_coherent_ep(ep)
            else:
                future_ep = _find_multi_input_ep(ep)
            if future_ep is None or self.depth[future_ep] - self.depth[sp] > limit:
                return self.create_block(sp, ep)
            ep = future_ep
        return self.create_block(sp=sp, ep=ep)

    def initialize_depth(self) -> None:
        """为图中所有节点确定深度，基于拓扑排序与动态规划，O(kn)时间复杂度 k为图节点最大度数，n为图节点个数."""
        for operation in self.op_orders:
            # graph input operation, set depth as 0
            if len(self.graph.get_upstream_operations(operation)) == 0:
                self.depth[operation] = 0
                continue

            # otherwise we will go dp
            depths_cache = []
            for up_op in self.graph.get_upstream_operations(operation):
                assert up_op in self.depth, 'Oops, that should not happen to your network.'
                depths_cache.append(self.depth[up_op])
            self.depth[operation] = max(depths_cache) + 1


class LSQDelegator(TorchQuantizeDelegator):
    def __init__(
        self,
        config: TensorQuantizationConfig,
        var: Variable,
        is_parameter_trainable: bool = True,
        is_scale_trainable: bool = True,
        is_offset_trainable: bool = True,
    ) -> None:
        self.config = config
        self.is_parameter = var.is_parameter
        self.var = var
        self.policy = config.policy
        self.passive = config.state == QuantizationStates.PASSIVE

        self.param_backup = None
        if self.is_parameter and is_parameter_trainable:
            self.param_backup = self.var.value.clone()

        # There is 4 checks for scale training:
        #   1. scale is valid
        #   2. state is active
        #   3. do not have POWER_OF_2 policy but Must have Linear policy
        #   4. is_scale_trainable = True
        self.scale_backup = None
        self.is_scale_trainable = False
        if is_scale_trainable:
            policy_check = not config.policy.has_property(QuantizationProperty.POWER_OF_2)
            linear_check = config.policy.has_property(QuantizationProperty.LINEAR)
            state_check = (config.state == QuantizationStates.ACTIVATED) and (config.dominated_by == config)
            value_check = isinstance(config.scale, torch.Tensor)
            if policy_check and state_check and value_check and linear_check:
                self.is_scale_trainable = True
                self.scale_backup = self.config.scale.detach().clone()

        # There is 4 checks for offset training:
        #   1. offset is valid
        #   2. state is active
        #   3. do not have SYMMETRICAL policy
        #   4. is_scale_trainable = True
        self.offset_backup = None
        self.is_offset_trainable = False
        if is_offset_trainable:
            policy_check = not config.policy.has_property(QuantizationProperty.SYMMETRICAL)
            state_check = (config.state == QuantizationStates.ACTIVATED) and (config.dominated_by == config)
            value_check = isinstance(config.offset, torch.Tensor)
            if policy_check and state_check and value_check:
                self.is_offset_trainable = True
                self.offset_backup = self.config.offset.detach().clone()

    def trainable_tensors(self) -> List[torch.Tensor]:
        params = []
        if self.is_offset_trainable:
            params.append(self.config.offset)
        if self.is_scale_trainable:
            params.append(self.config.scale)
        if self.is_parameter:
            params.append(self.var.value)
        return params

    def withdraw(self) -> None:
        with torch.no_grad():
            if self.scale_backup is not None:
                self.config.scale.copy_(self.scale_backup)
            if self.offset_backup is not None:
                self.config.offset.copy_(self.offset_backup)
            if self.param_backup is not None:
                self.var.value.copy_(self.param_backup)

    def finalize(self) -> None:
        self.scale_backup = None
        self.offset_backup = None
        self.param_backup = None
        pass  # do nothing here.

    def __call__(self, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        if tensor.is_cuda and PPQ_CONFIG.USING_CUDA_KERNEL:
            if config.policy.has_property(QuantizationProperty.LINEAR):
                if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                    return CuLSQ_LC.apply(
                        tensor,
                        config.scale,
                        config.offset,
                        config.channel_axis,
                        config.quant_min,
                        config.quant_max,
                        config.rounding,
                    )
                elif config.policy.has_property(QuantizationProperty.PER_TENSOR):
                    return CuLSQ_LT.apply(
                        tensor, config.scale, config.offset, config.quant_min, config.quant_max, config.rounding
                    )

            elif config.policy.has_property(QuantizationProperty.FLOATING):
                # For floating quantization, scale is not trainable.
                return PPQuantFunction(tensor=tensor, config=config)

        else:
            scale, offset = config.scale, config.offset

            if self.is_scale_trainable:
                scale = scale.abs()
                grad_scale = 1 / (tensor.numel() * config.quant_max) ** 0.5
                scale = scale * grad_scale + (scale - scale * grad_scale).detach()

            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                shape = [1 if axis != config.channel_axis else -1 for axis in range(tensor.ndim)]
                scale = scale.view(shape)
                offset = offset.view(shape)

            quantized = ppq_tensor_round((tensor / scale), config.rounding) + offset.detach()
            quantized = torch.clamp(quantized, config.quant_min, config.quant_max)
            quantized = (quantized - offset.detach()) * scale
            quantized = quantized
            return quantized


class RoundTuningDelegator(TorchQuantizeDelegator):
    def __init__(
        self,
        config: TensorQuantizationConfig,
        var: Variable,
    ) -> None:
        self.config = config
        self.is_parameter = var.is_parameter
        self.var = var
        self.policy = config.policy
        self.tunable_round = self.var.value

    def trainable_tensors(self) -> List[torch.Tensor]:
        params = []
        if self.is_parameter:
            params.append(self.var.value)
        return params

    def withdraw(self) -> None:
        with torch.no_grad():
            if self.scale_backup is not None:
                self.config.scale.copy_(self.scale_backup)
            if self.offset_backup is not None:
                self.config.offset.copy_(self.offset_backup)
            if self.param_backup is not None:
                self.var.value.copy_(self.param_backup)

    def finalize(self) -> None:
        self.scale_backup = None
        self.offset_backup = None
        self.param_backup = None
        pass  # do nothing here.

    def __call__(self, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        if tensor.is_cuda and PPQ_CONFIG.USING_CUDA_KERNEL:
            if config.policy.has_property(QuantizationProperty.LINEAR):
                if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                    return CuLSQ_LC.apply(
                        tensor,
                        config.scale,
                        config.offset,
                        config.channel_axis,
                        config.quant_min,
                        config.quant_max,
                        config.rounding,
                    )
                elif config.policy.has_property(QuantizationProperty.PER_TENSOR):
                    return CuLSQ_LT.apply(
                        tensor, config.scale, config.offset, config.quant_min, config.quant_max, config.rounding
                    )

            elif config.policy.has_property(QuantizationProperty.FLOATING):
                # For floating quantization, scale is not trainable.
                return PPQuantFunction(tensor=tensor, config=config)

        else:
            scale, offset = config.scale, config.offset

            if self.is_scale_trainable:
                scale = scale.abs()
                grad_scale = 1 / (tensor.numel() * config.quant_max) ** 0.5
                scale = scale * grad_scale + (scale - scale * grad_scale).detach()

            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                shape = [1 if axis != config.channel_axis else -1 for axis in range(tensor.ndim)]
                scale = scale.view(shape)
                offset = offset.view(shape)

            quantized = ppq_tensor_round((tensor / scale), config.rounding) + offset.detach()
            quantized = torch.clamp(quantized, config.quant_min, config.quant_max)
            quantized = (quantized - offset.detach()) * scale
            quantized = quantized
            return quantized


class TensorwiseRoundTuningImpl(Function):
    @staticmethod
    def forward(
        ctx,
        tensor: torch.Tensor,
        scale: torch.Tensor,
        offset: torch.Tensor,
        quant_min: int,
        quant_max: int,
        rounding: torch.Tensor,
    ) -> torch.Tensor:
        scale, offset = scale.to(tensor.device), offset.to(tensor.device)
        tensor = (tensor / scale) + (rounding > 0.5) + offset
        tensor = torch.clamp(tensor, quant_min, quant_max)
        tensor = (tensor - offset) * scale
        return tensor

    @staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None, None, None, None, dy


class ChannelwiseRoundTuningImpl(Function):
    @staticmethod
    def forward(
        ctx,
        tensor: torch.Tensor,
        scale: torch.Tensor,
        offset: torch.Tensor,
        channel_axis: int,
        quant_min: int,
        quant_max: int,
        rounding: torch.Tensor,
    ) -> torch.Tensor:
        scale, offset = scale.to(tensor.device), offset.to(tensor.device)
        # generate a shape that likes [1, 1, -1, 1], the only -1 is at channel axe.
        shape = [1 if axis != channel_axis else -1 for axis in range(tensor.ndim)]
        scale, offset = scale.view(shape), offset.view(shape)

        tensor = (tensor / scale) + (rounding > 0.5) + offset
        tensor = torch.clamp(tensor, quant_min, quant_max)
        tensor = (tensor - offset) * scale
        return tensor

    @staticmethod
    def backward(ctx, dy: torch.Tensor):
        return dy, None, None, None, None, None, dy


class RoundTruningDelegator(TorchQuantizeDelegator):
    def __init__(
        self,
        var: Variable,
        config: TensorQuantizationConfig,
    ) -> None:
        self.config = config
        self.var = var
        self.is_parameter = self.var.is_parameter

        # environment check
        if config.policy.has_property(QuantizationProperty.FLOATING):
            raise TypeError('Incorrect Quantization Property. Except Linear Quantization Policy.')
        if config.policy.has_property(QuantizationProperty.DYNAMIC):
            raise TypeError('Incorrect Quantization Property. Except Static Quantization Policy.')
        if not self.var.is_parameter:
            raise TypeError(f'Variable {self.var.name} is not a parameter!')
        if self.var.value is None or not isinstance(self.var.value, torch.Tensor):
            raise ValueError(f'Unexpected value type of {self.var.name}')
        if self.config.scale is None:
            raise ValueError('Quantization Scale has not been correctly set.')

        # initialize rounding
        self._calling_times = 0
        self._executing_device = self.var.value.device
        self._param_backup = self.var.value.clone()

        with torch.no_grad():
            scale, _ = config.scale, config.offset
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                shape = [1 if axis != config.channel_axis else -1 for axis in range(self.var.value.ndim)]
                scale = scale.view(shape)

            rounding = (self.var.value / scale) - (self.var.value / scale).floor()
            self.var.value = (self.var.value / scale).floor() * scale
            self._scale = scale

        # create grad.
        self._rounding = rounding
        self._rounding.requires_grad = True

    def trainable_tensors(self) -> List[torch.Tensor]:
        return [self._rounding]

    def finalize(self) -> None:
        with torch.no_grad():
            self.var.value += (self._rounding > 0.5) * self._scale

    def withdraw(self) -> None:
        with torch.no_grad():
            self.var.value.copy_(self._param_backup)

    def __call__(self, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        if config.policy.has_property(QuantizationProperty.PER_TENSOR):
            return TensorwiseRoundTuningImpl.apply(
                tensor, config.scale, config.offset, config.quant_min, config.quant_max, self._rounding
            )
        elif config.policy.has_property(QuantizationProperty.PER_CHANNEL):
            return ChannelwiseRoundTuningImpl.apply(
                tensor,
                config.scale,
                config.offset,
                config.channel_axis,
                config.quant_min,
                config.quant_max,
                self._rounding,
            )
        else:
            raise Exception('Oops, this should not happen.')



# Test Quantization System Performace on OCR Models with IC15 Dataset
#
#   1. How to use:
#      Run this script with python directly.
#

# Quantizer Configuration
SYMMETRICAL = True
PERCHANNEL = True
POWER_OF_2 = False
FP8 = False
BIT_WIDTH = 8

# Should contains model file(.onnx)
MODEL_DIR = 'QuantZoo/Model/ocr'

# Should contains Calib & Test Img Folder
CALIB_DIR = 'QuantZoo/Data/IC15'
CALIB_LABEL = 'QuantZoo/Data/IC15/rec_gt_train.txt'
TEST_DIR = 'QuantZoo/Data/IC15'
TEST_LABEL = 'QuantZoo/Data/IC15/rec_gt_test.txt'
CHAR_DIR = 'QuantZoo/Data/IC15/ic15_dict.txt'

# calibration & test batchsize
BATCHSIZE = 32

# write report to here
REPORT_DIR = 'QuantZoo/Reports'

CONFIGS = [
    {
        'Model': 'en_PP-OCRv3_rec_infer',
        'Output': ['swish_13.tmp_0'],
        'Dictionary': 'en_dict.txt',
        'Reshape': [3, 48, 320],
        'Language': 'en',
    },
    {
        'Model': 'en_number_mobile_v2.0_rec_infer',
        'Output': ['save_infer_model/scale_0.tmp_1'],
        'Dictionary': 'en_dict.txt',
        'Reshape': [3, 32, 320],
        'Language': 'en',
    },
    {
        'Model': 'ch_PP-OCRv2_rec_infer',
        'Output': ['p2o.LSTM.5'],
        'Dictionary': 'ppocr_keys_v1.txt',
        'Reshape': [3, 32, 320],
        'Language': 'ch',
    },
    {
        'Model': 'ch_PP-OCRv3_rec_infer',
        'Output': ['swish_27.tmp_0'],
        'Dictionary': 'ppocr_keys_v1.txt',
        'Reshape': [3, 48, 320],
        'Language': 'ch',
    },
    {
        'Model': 'ch_ppocr_mobile_v2.0_rec_infer',
        'Output': ['p2o.LSTM.5'],
        'Dictionary': 'ppocr_keys_v1.txt',
        'Reshape': [3, 32, 320],
        'Language': 'ch',
    },
    {
        'Model': 'ch_ppocr_server_v2.0_rec_infer',
        'Output': ['p2o.LSTM.5'],
        'Dictionary': 'ppocr_keys_v1.txt',
        'Reshape': [3, 32, 320],
        'Language': 'ch',
    },
]

import os

import torch
from QuantZoo.Data.IC15.Data import IC15_PaddleOCR
from QuantZoo.Data.IC15.Eval import evaluate_ppq_module_with_ic15
from QuantZoo.Quantizers import MyFP8Quantizer, MyInt8Quantizer
from QuantZoo.Util import error_analyze, report

import esp_ppq.lib as PFL
from esp_ppq import convert_any_to_torch_tensor
from esp_ppq.api import ENABLE_CUDA_KERNEL, load_onnx_graph
from esp_ppq.core import TargetPlatform
from esp_ppq.executor import TorchExecutor
from esp_ppq.quantization.optim import (
    LayerwiseEqualizationPass,
    LearnedStepSizePass,
    ParameterQuantizePass,
    RuntimeCalibrationPass,
)

with ENABLE_CUDA_KERNEL():
    for config in CONFIGS:
        model = config['Model']
        monitoring_vars = config['Output']
        dictionary = config['Dictionary']
        shape = config['Reshape']
        chinese = config['Language'] == 'ch'

        calib_loader = IC15_PaddleOCR(
            images_path=CALIB_DIR, label_path=CALIB_LABEL, input_shape=shape, is_chinese_version=chinese
        ).dataloader(batchsize=BATCHSIZE, shuffle=False)

        test_loader = IC15_PaddleOCR(
            images_path=TEST_DIR, label_path=TEST_LABEL, input_shape=shape, is_chinese_version=chinese
        ).dataloader(batchsize=BATCHSIZE, shuffle=False)

        print(f"Ready to run quant benchmark on {model}")
        graph = load_onnx_graph(onnx_import_file=os.path.join(MODEL_DIR, model + '.onnx'))

        quantizer = MyInt8Quantizer(
            graph=graph, sym=SYMMETRICAL, power_of_2=POWER_OF_2, num_of_bits=BIT_WIDTH, per_channel=PERCHANNEL
        )
        if FP8:
            quantizer = MyFP8Quantizer(graph=graph, calibration='floating')

        # convert op to quantable-op
        for name, op in graph.operations.items():
            if op.type in {'Conv', 'ConvTranspose', 'MatMul', 'Gemm'}:
                quantizer.quantize_operation(name, platform=TargetPlatform.INT8)

        # build quant pipeline.
        pipeline = PFL.Pipeline(
            [
                # LayerwiseEqualizationPass(iteration=10),
                ParameterQuantizePass(),
                RuntimeCalibrationPass(),
                # LearnedStepSizePass(steps=500, collecting_device='cuda')
            ]
        )

        # call pipeline.
        executor = TorchExecutor(graph=graph)
        executor.tracing_operation_meta(torch.zeros(size=[BATCHSIZE, 3, 32, 100]).cuda())

        pipeline.optimize(
            graph=graph,
            dataloader=calib_loader,
            verbose=True,
            calib_steps=32,
            collate_fn=lambda x: x[0].to('cuda'),
            executor=executor,
        )

        acc = evaluate_ppq_module_with_ic15(
            executor=executor,
            character_dict_path=os.path.join(MODEL_DIR, dictionary),
            dataloader=test_loader,
            collate_fn=lambda x: convert_any_to_torch_tensor(x).cuda(),
        )
        print(f'Model Performace on IC15: {acc * 100:.4f}%')

        # error analyze
        performance = error_analyze(
            graph=graph,
            outputs=monitoring_vars,
            dataloader=test_loader,
            collate_fn=lambda x: x[0].to('cuda'),
            verbose=True,
        )



import bisect
from typing import List

import numpy as np
import torch

from esp_ppq.core import (
    DataType,
    OperationQuantizationConfig,
    convert_any_to_torch_tensor,
)
from esp_ppq.IR import BaseGraph, Operation, OperationExporter, Variable
from esp_ppq.IR.quantize import QuantableOperation
from esp_ppq.log import NaiveLogger
from esp_ppq.parser.espdl.espdl_graph_utils import (
    fuse_downstream_operation,
    get_default_perm,
    get_inverse_transpose,
    insert_transpose_node,
    restore_origin_shape,
    transpose_shape,
)
from esp_ppq.parser.espdl.espdl_typedef import (
    ADD_LIKE_OP_SET,
    AXIS_TRANSFORM_OP_SET,
    CONV_LAYOUT_OP_SET,
    OTHER_OP_SET,
    PASSIVE_LAYOUT_OP_SET,
    REDUCE_OP_SET,
    SOFTMAX_LIKE_OP_SET,
    ExporterPatternInfo,
)

logger = NaiveLogger.get_logger('ESPDL')
# logger.set_level("DEBUG")


class ResetConvLayoutPattern(OperationExporter):
    """
    Modify Conv inputs and outputs layout from NCHW to NHWC
    And Update all variable's shape
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        info = ExporterPatternInfo()
        if op.type in CONV_LAYOUT_OP_SET:
            for var in op.inputs:
                if var.is_parameter:
                    continue

                var_shape = var.shape
                if len(var_shape) == 4:  # conv2d, NCHW -> NHWC
                    perm = [0, 2, 3, 1]
                elif len(var_shape) == 3:  # conv1d, NCW -> NWC
                    perm = [0, 2, 1]

                var_perm = info.get_var_permute(var.name)

                if var_perm:
                    if perm != var_perm:
                        # There is already a permute, but it does not match the conv layout.
                        # A transpose node needs to be inserted into the graph.
                        inverse_perm = get_inverse_transpose(var_perm)
                        new_perm = transpose_shape(inverse_perm, perm)
                        insert_transpose_node(graph, var, op, new_perm)
                else:
                    info.add_var_permute(var.name, perm)

            for var in op.outputs:
                var_shape = var.shape
                if len(var_shape) == 4:  # conv2d, NCHW -> NHWC
                    perm = [0, 2, 3, 1]
                else:  # conv1d, NCW -> NWC
                    perm = [0, 2, 1]
                info.add_var_permute(var.name, perm)

        return op


class RestoreOriginLayoutPattern(OperationExporter):
    """
    Restore original layout
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in OTHER_OP_SET:
            return restore_origin_shape(op, graph)

        return op


class BypassPassiveLayoutPattern(OperationExporter):
    """
    Passive Node inherit transpose from upstream
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in PASSIVE_LAYOUT_OP_SET:
            info = ExporterPatternInfo()
            assert len(op.outputs) == 1
            input1 = op.inputs[0]
            output = op.outputs[0]
            assert input1.shape == output.shape

            var_perm = info.get_var_permute(input1.name)
            if not var_perm:
                var_perm = get_default_perm(input1)
                info.add_var_permute(input1.name, var_perm)

            info.add_var_permute(op.outputs[0].name, var_perm)
        return op


class BypassAddLikePattern(OperationExporter):
    """
    Add,Mul,Sub,Div:

    two input and one output,
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in ADD_LIKE_OP_SET:
            info = ExporterPatternInfo()
            input1 = op.inputs[0]
            input2 = op.inputs[1]
            output = op.outputs[0]
            input1_perm = info.get_var_permute(input1.name)
            input2_perm = info.get_var_permute(input2.name)
            output_perm = None
            if not input1.is_parameter and not input2.is_parameter:
                if input1_perm == input2_perm:
                    if input1_perm:
                        # using upstream's perm
                        output_perm = input1_perm
                    else:
                        # input1_perm is None, add new perm
                        info.add_var_permute(input1.name, get_default_perm(input1))
                        info.add_var_permute(input2.name, get_default_perm(input2))
                        output_perm = get_default_perm(output)
                    # logger.debug(f"{info.get_var_permute(input1.name)}, {info.get_var_permute(input2.name)}, {output_perm}")
                    info.add_var_permute(output.name, output_perm)
                else:
                    # insert transpose node and restore origin shape
                    return restore_origin_shape(op, graph)
            elif input2.is_parameter or input1.is_parameter:
                return restore_origin_shape(op, graph)

        return op


class AxisTransformPattern(OperationExporter):
    """
    Transform the axes of operator
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        info = ExporterPatternInfo()
        input = op.inputs[0]

        var_perm = info.get_var_permute(input.name)
        if var_perm and var_perm != get_default_perm(input):
            # There is already a permute, change axis accordingly
            if op.type in SOFTMAX_LIKE_OP_SET:
                axis = (int(op.attributes["axis"]) + len(var_perm)) % len(var_perm)
                new_axis = var_perm.index(axis)
                op.attributes["axis"] = new_axis
            elif op.type in REDUCE_OP_SET:
                if len(op.inputs) > 1 and len(op.inputs[1].value) > 0:
                    axes = op.inputs[1].value
                    new_axes = []
                    for i in range(axes.numel()):
                        if axes[i] < 0:
                            new_axes.append(var_perm.index(axes[i] + len(input.shape)))
                        else:
                            new_axes.append(var_perm.index(axes[i]))
                    new_axes = sorted(new_axes)
                    op.inputs[1].value = convert_any_to_torch_tensor(new_axes, dtype=torch.int64)
        else:
            var_perm = get_default_perm(input)
            info.add_var_permute(input.name, var_perm)

        out_var_perm = var_perm[:]
        if op.type in REDUCE_OP_SET:
            keepdims = op.attributes["keepdims"]
            # The perm needs to be deleted.
            if keepdims == 0:
                if len(op.inputs) > 1 and len(op.inputs[1].value) > 0:
                    reduce_axes_list = sorted(op.inputs[1].value.tolist(), reverse=True)
                    # After the transformations above, the axes now match the current input.
                    # Perm represents the current state of the input, not the permutation about to be transformed.
                    for axis in reduce_axes_list:
                        remove_axis = out_var_perm.pop(axis)
                        for i in range(len(out_var_perm)):
                            if out_var_perm[i] > remove_axis:
                                out_var_perm[i] -= 1
                    if len(out_var_perm) == 0:
                        out_var_perm = [0]
                else:
                    noop_with_empty_axes = bool(op.attributes.get('noop_with_empty_axes', 0))
                    if not noop_with_empty_axes:
                        out_var_perm = [0]

        for output in op.outputs:
            info.add_var_permute(output.name, out_var_perm)

        return op


class ResetConcatPattern(OperationExporter):
    """
    Concat pattern with two input and one output,
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in ["Concat"]:
            perm_dict = {}
            info = ExporterPatternInfo()

            for var in op.inputs:
                var_perm = info.get_var_permute(var.name)
                perm_str = str(var_perm)
                if perm_str not in perm_dict:
                    perm_dict[perm_str] = var_perm

            output_var = op.outputs[0]

            if len(perm_dict) == 1:  # all input have same perm, output bypass
                var_perm = list(perm_dict.values())[0]
                if not var_perm:
                    restore_origin_shape(op, graph)
                else:
                    axis = op.attributes["axis"]
                    new_axis = var_perm.index(int(axis))
                    op.attributes["axis"] = new_axis
                    info.add_var_permute(output_var.name, var_perm)
                    logger.debug(f"{op.name} update axes from {axis} to {new_axis}")
            else:
                logger.debug(f"transpose perm {perm_dict}")
                restore_origin_shape(op, graph)
        return op


class ResetResizePattern(OperationExporter):
    """
    Reize Layout Pattern
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in ["Resize"]:
            info = ExporterPatternInfo()
            input_var = op.inputs[0]
            input_shape_orig = input_var.shape
            perm = None
            if len(input_shape_orig) == 4:  # 2d, NCHW -> NHWC
                perm = [0, 2, 3, 1]
            elif len(input_shape_orig) == 3:  # 1d, NCW -> NWC
                perm = [0, 2, 1]
            else:
                logger.error(f"Reize: do not support shape for {input_shape_orig}")

            input_perm = info.get_var_permute(input_var.name)
            if input_perm:
                if perm != input_perm:
                    # There is already a permute, but it does not match the Reize layout.
                    # A transpose node needs to be inserted into the graph.
                    inverse_perm = get_inverse_transpose(input_perm)
                    new_perm = transpose_shape(inverse_perm, perm)
                    insert_transpose_node(graph, input_var, op, new_perm)
            else:
                info.add_var_permute(input_var.name, perm)

            # Get roi
            roi_var = op.inputs[1]
            if not roi_var.name or roi_var.value.numel() == 0:
                roi_var = None

            # Get scales
            scales_var = op.inputs[2]
            if not scales_var.name:
                scales_var = None

            # Get sizes
            if len(op.inputs) > 3:
                sizes_var = op.inputs[3]
            else:
                sizes_var = None

            if scales_var is None and sizes_var is None:
                raise ValueError("scales_var is None and sizes_var is None.")

            rank = len(input_shape_orig)
            axes = op.attributes.get("axes", None)

            # Adjust roi, scales, and sizes to the same dimensions as input.
            # But the data arrangement is still in NCHW or NCW format.
            if axes is not None:
                # Adjust roi parameters
                if roi_var is not None:
                    new_roi_value = ([0.0] * rank) + ([1.0] * rank)
                    naxes = len(axes)
                    for i, d in enumerate(axes):
                        new_roi_value[d] = roi_var.value[i]
                        new_roi_value[rank + d] = roi_var.value[naxes + i]
                    roi_var.value = torch.tensor(new_roi_value).type(roi_var.dtype.to_torch())
                    roi_var.shape = roi_var.value.shape

                # Adjust scales parameters
                if scales_var is not None:
                    new_scales_value = [1.0] * rank
                    for i, d in enumerate(axes):
                        new_scales_value[d] = scales_var.value[i]
                    scales_var.value = torch.tensor(new_scales_value).type(scales_var.dtype.to_torch())
                    scales_var.shape = scales_var.value.shape

                # Adjust sizes parameters
                if sizes_var is not None:
                    new_sizes_value = [input_shape_orig[i] for i in range(rank)]
                    for i, d in enumerate(axes):
                        new_sizes_value[d] = sizes_var.value[i]
                    sizes_var.value = torch.tensor(new_sizes_value).type(sizes_var.dtype.to_torch())
                    sizes_var.shape = sizes_var.value.shape

            else:
                axes = list(range(rank))

            # Adjust scales and sizes according to keep_aspect_ratio_policy.
            # This attribute describes how to interpret the sizes input with regard to keeping the original
            # aspect ratio of the input, and it is not applicable when the scales input is used.
            keep_aspect_ratio_policy = op.attributes.get("keep_aspect_ratio_policy", None)
            if sizes_var is not None:
                scale_factors = [sizes_var.value[i] / input_shape_orig[i] for i in range(rank)]
                if keep_aspect_ratio_policy and keep_aspect_ratio_policy != "stretch":
                    if keep_aspect_ratio_policy == "not_larger":
                        scale = np.array(scale_factors)[axes].min()
                    elif keep_aspect_ratio_policy == "not_smaller":
                        scale = np.array(scale_factors)[axes].max()
                    else:
                        raise ValueError(f"Invalid keep_aspect_ratio_policy={keep_aspect_ratio_policy!r}")

                    scale_factors = [scale if i in axes else 1.0 for i in range(rank)]

                    def round_half_up(x: float) -> int:
                        return int(x + 0.5)

                    output_size = [
                        round_half_up(scale * input_shape_orig[i]) if i in axes else input_shape_orig[i]
                        for i in range(rank)
                    ]
                else:
                    output_size = sizes_var.value.tolist()

                if scales_var:
                    scales_var.value = torch.tensor(scale_factors).type(dtype=DataType.to_torch(scales_var.dtype))
            else:
                output_size = (scales_var.value * torch.tensor(input_shape_orig)).type(torch.int64)  # type: ignore[union-attr]

            if sizes_var:
                sizes_var.value = torch.tensor(output_size).type(dtype=DataType.to_torch(sizes_var.dtype))

            # Align the data arrangement of the parameters with that of the input.
            # if perm:
            #     if roi_var is not None:
            #         roi_value_1 = torch.tensor(transpose_shape(roi_var.value[0:rank], perm)).type(dtype=DataType.to_torch(roi_var.dtype))
            #         roi_value_2 = torch.tensor(transpose_shape(roi_var.value[rank:], perm)).type(dtype=DataType.to_torch(roi_var.dtype))
            #         roi_var.value = torch.cat((roi_value_1, roi_value_2))
            #         roi_var.shape = roi_var.value.shape
            #     if scales_var is not None:
            #         scales_var.value = torch.tensor(transpose_shape(scales_var.value, perm)).type(dtype=DataType.to_torch(scales_var.dtype))
            #         scales_var.shape = scales_var.value.shape
            #     if sizes_var is not None:
            #         sizes_var.value = torch.tensor(transpose_shape(sizes_var.value, perm)).type(dtype=DataType.to_torch(sizes_var.dtype))
            #         sizes_var.shape = sizes_var.value.shape

            for var in op.inputs[1:]:
                if var:
                    info.add_var_permute(var.name, get_default_perm(var))
            info.add_var_permute(op.outputs[0].name, perm)

        return op


class FuseTransposePattern(OperationExporter):
    """
    Fuse Transpose Pattern
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        # The FUSE_OP_PATTERNS may remove some ops.
        if op.name not in graph.operations:
            return op

        if op.type == "Transpose":
            downstream_transpose_op = None
            perm = op.attributes["perm"]
            while True:
                downstream_op = graph.get_downstream_operations(op)
                # the downstream op have only one op and this op is Transpose
                if len(downstream_op) == 1 and downstream_op[0].type == "Transpose":
                    downstream_transpose_op = downstream_op[0]
                    perm = transpose_shape(perm, downstream_op[0].attributes["perm"])

                    if isinstance(op, QuantableOperation):
                        op_config = op.config
                        downstream_config = downstream_transpose_op.config
                        new_config = OperationQuantizationConfig(
                            op_config.input_quantization_config,
                            downstream_config.output_quantization_config,
                        )
                        op.config = new_config
                    graph = fuse_downstream_operation(graph, downstream_transpose_op, keep_coherence=True)
                    op.attributes["perm"] = perm
                else:
                    break

            perm = op.attributes["perm"]
            if perm == [i for i in range(len(perm))]:
                # Removed redundant transpose
                graph.remove_operation(op, keep_coherence=True)
        return op


def print_vars(op: Operation):
    logger.info(f"Op: {op.name}, {op.type}, {op.attributes}")
    for var in op.inputs:
        print("inputs:", var.name, var.shape)
    for var in op.outputs:
        print("outputs:", var.name, var.shape)


def reset_graph_layout(graph: BaseGraph):
    """
    Reset layout from NCHW -> NHWC
    """

    layout_patterns = [
        [CONV_LAYOUT_OP_SET, ResetConvLayoutPattern],
        [PASSIVE_LAYOUT_OP_SET, BypassPassiveLayoutPattern],
        [ADD_LIKE_OP_SET, BypassAddLikePattern],
        [AXIS_TRANSFORM_OP_SET, AxisTransformPattern],
        [["Concat"], ResetConcatPattern],
        [["Resize"], ResetResizePattern],
        [OTHER_OP_SET, RestoreOriginLayoutPattern],
    ]

    for op in graph.topological_sort():
        flag = 1
        for pattern in layout_patterns:
            if op.type in pattern[0]:
                pattern[1]().export(op, graph)
                flag = 0
                break
        if flag:
            logger.error(f"Can not reset {op.type}:{op.name} layout")

    # fuse transpose op
    pattern = FuseTransposePattern()
    for op in graph.topological_sort():
        pattern.export(op, graph)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Function(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Function()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsFunction(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Function
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Function
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Function
    def Input(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Function
    def InputLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Function
    def InputIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Function
    def Output(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Function
    def OutputLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Function
    def OutputIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Function
    def Attribute(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Function
    def AttributeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Function
    def AttributeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Function
    def AttributeProto(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.Attribute import Attribute
            obj = Attribute()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Function
    def AttributeProtoLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Function
    def AttributeProtoIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Function
    def Node(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.Node import Node
            obj = Node()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Function
    def NodeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Function
    def NodeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Function
    def DocString(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Function
    def OpsetImport(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.OperatorSetId import OperatorSetId
            obj = OperatorSetId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Function
    def OpsetImportLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Function
    def OpsetImportIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # Function
    def Domain(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def FunctionStart(builder):
    builder.StartObject(9)

def Start(builder):
    FunctionStart(builder)

def FunctionAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)

def AddName(builder, name):
    FunctionAddName(builder, name)

def FunctionAddInput(builder, input):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(input), 0)

def AddInput(builder, input):
    FunctionAddInput(builder, input)

def FunctionStartInputVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartInputVector(builder, numElems):
    return FunctionStartInputVector(builder, numElems)

def FunctionAddOutput(builder, output):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(output), 0)

def AddOutput(builder, output):
    FunctionAddOutput(builder, output)

def FunctionStartOutputVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartOutputVector(builder, numElems):
    return FunctionStartOutputVector(builder, numElems)

def FunctionAddAttribute(builder, attribute):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(attribute), 0)

def AddAttribute(builder, attribute):
    FunctionAddAttribute(builder, attribute)

def FunctionStartAttributeVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartAttributeVector(builder, numElems):
    return FunctionStartAttributeVector(builder, numElems)

def FunctionAddAttributeProto(builder, attributeProto):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(attributeProto), 0)

def AddAttributeProto(builder, attributeProto):
    FunctionAddAttributeProto(builder, attributeProto)

def FunctionStartAttributeProtoVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartAttributeProtoVector(builder, numElems):
    return FunctionStartAttributeProtoVector(builder, numElems)

def FunctionAddNode(builder, node):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(node), 0)

def AddNode(builder, node):
    FunctionAddNode(builder, node)

def FunctionStartNodeVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartNodeVector(builder, numElems):
    return FunctionStartNodeVector(builder, numElems)

def FunctionAddDocString(builder, docString):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(docString), 0)

def AddDocString(builder, docString):
    FunctionAddDocString(builder, docString)

def FunctionAddOpsetImport(builder, opsetImport):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(opsetImport), 0)

def AddOpsetImport(builder, opsetImport):
    FunctionAddOpsetImport(builder, opsetImport)

def FunctionStartOpsetImportVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartOpsetImportVector(builder, numElems):
    return FunctionStartOpsetImportVector(builder, numElems)

def FunctionAddDomain(builder, domain):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(domain), 0)

def AddDomain(builder, domain):
    FunctionAddDomain(builder, domain)

def FunctionEnd(builder):
    return builder.EndObject()

def End(builder):
    return FunctionEnd(builder)

import FlatBuffers.Dl.Attribute
import FlatBuffers.Dl.Node
import FlatBuffers.Dl.OperatorSetId
try:
    from typing import List
except:
    pass

class FunctionT(object):

    # FunctionT
    def __init__(self):
        self.name = None  # type: str
        self.input = None  # type: List[str]
        self.output = None  # type: List[str]
        self.attribute = None  # type: List[str]
        self.attributeProto = None  # type: List[FlatBuffers.Dl.Attribute.AttributeT]
        self.node = None  # type: List[FlatBuffers.Dl.Node.NodeT]
        self.docString = None  # type: str
        self.opsetImport = None  # type: List[FlatBuffers.Dl.OperatorSetId.OperatorSetIdT]
        self.domain = None  # type: str

    @classmethod
    def InitFromBuf(cls, buf, pos):
        function = Function()
        function.Init(buf, pos)
        return cls.InitFromObj(function)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, function):
        x = FunctionT()
        x._UnPack(function)
        return x

    # FunctionT
    def _UnPack(self, function):
        if function is None:
            return
        self.name = function.Name()
        if not function.InputIsNone():
            self.input = []
            for i in range(function.InputLength()):
                self.input.append(function.Input(i))
        if not function.OutputIsNone():
            self.output = []
            for i in range(function.OutputLength()):
                self.output.append(function.Output(i))
        if not function.AttributeIsNone():
            self.attribute = []
            for i in range(function.AttributeLength()):
                self.attribute.append(function.Attribute(i))
        if not function.AttributeProtoIsNone():
            self.attributeProto = []
            for i in range(function.AttributeProtoLength()):
                if function.AttributeProto(i) is None:
                    self.attributeProto.append(None)
                else:
                    attribute_ = FlatBuffers.Dl.Attribute.AttributeT.InitFromObj(function.AttributeProto(i))
                    self.attributeProto.append(attribute_)
        if not function.NodeIsNone():
            self.node = []
            for i in range(function.NodeLength()):
                if function.Node(i) is None:
                    self.node.append(None)
                else:
                    node_ = FlatBuffers.Dl.Node.NodeT.InitFromObj(function.Node(i))
                    self.node.append(node_)
        self.docString = function.DocString()
        if not function.OpsetImportIsNone():
            self.opsetImport = []
            for i in range(function.OpsetImportLength()):
                if function.OpsetImport(i) is None:
                    self.opsetImport.append(None)
                else:
                    operatorSetId_ = FlatBuffers.Dl.OperatorSetId.OperatorSetIdT.InitFromObj(function.OpsetImport(i))
                    self.opsetImport.append(operatorSetId_)
        self.domain = function.Domain()

    # FunctionT
    def Pack(self, builder):
        if self.name is not None:
            name = builder.CreateString(self.name)
        if self.input is not None:
            inputlist = []
            for i in range(len(self.input)):
                inputlist.append(builder.CreateString(self.input[i]))
            FunctionStartInputVector(builder, len(self.input))
            for i in reversed(range(len(self.input))):
                builder.PrependUOffsetTRelative(inputlist[i])
            input = builder.EndVector()
        if self.output is not None:
            outputlist = []
            for i in range(len(self.output)):
                outputlist.append(builder.CreateString(self.output[i]))
            FunctionStartOutputVector(builder, len(self.output))
            for i in reversed(range(len(self.output))):
                builder.PrependUOffsetTRelative(outputlist[i])
            output = builder.EndVector()
        if self.attribute is not None:
            attributelist = []
            for i in range(len(self.attribute)):
                attributelist.append(builder.CreateString(self.attribute[i]))
            FunctionStartAttributeVector(builder, len(self.attribute))
            for i in reversed(range(len(self.attribute))):
                builder.PrependUOffsetTRelative(attributelist[i])
            attribute = builder.EndVector()
        if self.attributeProto is not None:
            attributeProtolist = []
            for i in range(len(self.attributeProto)):
                attributeProtolist.append(self.attributeProto[i].Pack(builder))
            FunctionStartAttributeProtoVector(builder, len(self.attributeProto))
            for i in reversed(range(len(self.attributeProto))):
                builder.PrependUOffsetTRelative(attributeProtolist[i])
            attributeProto = builder.EndVector()
        if self.node is not None:
            nodelist = []
            for i in range(len(self.node)):
                nodelist.append(self.node[i].Pack(builder))
            FunctionStartNodeVector(builder, len(self.node))
            for i in reversed(range(len(self.node))):
                builder.PrependUOffsetTRelative(nodelist[i])
            node = builder.EndVector()
        if self.docString is not None:
            docString = builder.CreateString(self.docString)
        if self.opsetImport is not None:
            opsetImportlist = []
            for i in range(len(self.opsetImport)):
                opsetImportlist.append(self.opsetImport[i].Pack(builder))
            FunctionStartOpsetImportVector(builder, len(self.opsetImport))
            for i in reversed(range(len(self.opsetImport))):
                builder.PrependUOffsetTRelative(opsetImportlist[i])
            opsetImport = builder.EndVector()
        if self.domain is not None:
            domain = builder.CreateString(self.domain)
        FunctionStart(builder)
        if self.name is not None:
            FunctionAddName(builder, name)
        if self.input is not None:
            FunctionAddInput(builder, input)
        if self.output is not None:
            FunctionAddOutput(builder, output)
        if self.attribute is not None:
            FunctionAddAttribute(builder, attribute)
        if self.attributeProto is not None:
            FunctionAddAttributeProto(builder, attributeProto)
        if self.node is not None:
            FunctionAddNode(builder, node)
        if self.docString is not None:
            FunctionAddDocString(builder, docString)
        if self.opsetImport is not None:
            FunctionAddOpsetImport(builder, opsetImport)
        if self.domain is not None:
            FunctionAddDomain(builder, domain)
        function = FunctionEnd(builder)
        return function



from esp_ppq.core import NetworkFramework, TargetPlatform
from esp_ppq.IR import BaseGraph, GraphBuilder, GraphExporter

from .ascend_export import AscendExporter
from .espdl_exporter import EspdlExporter
from .extension import ExtensionExporter
from .mnn_exporter import MNNExporter
from .native import NativeExporter, NativeImporter
from .ncnn_exporter import NCNNExporter
from .nxp_exporter import NxpExporter
from .onnx_exporter import OnnxExporter
from .onnx_parser import OnnxParser
from .onnxruntime_exporter import ONNXRUNTIMExporter
from .openvino_exporter import OpenvinoExporter
from .ppl import PPLBackendExporter
from .qnn_exporter import QNNDSPExporter
from .tengine_exporter import TengineExporter
from .tensorRT import TensorRTExporter_JSON, TensorRTExporter_QDQ



import torch
import torchvision
from trainer import ImageNetTrainer
from Utilities.Imagenet import *  # check esp_ppq.samples.imagenet.Utilities
from Utilities.Imagenet.imagenet_util import load_imagenet_from_directory  # check esp_ppq.samples.imagenet.Utilities

import esp_ppq.lib as PFL
from esp_ppq.api import ENABLE_CUDA_KERNEL, load_native_graph, load_torch_model
from esp_ppq.core import (
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    QuantizationVisibility,
    RoundingPolicy,
    TargetPlatform,
)
from esp_ppq.executor import TorchExecutor
from esp_ppq.IR import Operation
from esp_ppq.quantization.optim import *

"""
    使用这个脚本来尝试在 Imagenet 数据集上执行量化感知训练
        使用 imagenet 中的数据测试量化精度与 calibration
        默认的 imagenet 数据集位置:Assets/Imagenet_Train, Assets/Imagenet_Valid
        你可以通过软连接创建它们:
            ln -s /home/data/Imagenet/val Assets/Imagenet_Valid
            ln -s /home/data/Imagenet/train Assets/Imagenet_Train
"""

CFG_DEVICE = 'cuda'  # 一个神奇的字符串，用来确定执行设备
CFG_BATCHSIZE = 32  # 测试与calib时的 batchsize
CFG_INPUT_SHAPE = (CFG_BATCHSIZE, 3, 224, 224)  # 用来确定模型输入的尺寸，好像 imagenet 都是这个尺寸
CFG_VALIDATION_DIR = 'Assets/Imagenet_Valid'  # 用来读取 validation dataset
CFG_TRAIN_DIR = 'Assets/Imagenet_Train'  # 用来读取 train dataset，注意该集合将被用来 calibrate 你的模型
CFG_PLATFORM = TargetPlatform.TRT_INT8  # 用来指定目标平台

# ------------------------------------------------------------
# 在这个例子中我们将向你展示如何在 PPQ 中对你的网络进行量化感知训练
# 你可以使用带标签的数据执行正常的训练流程，也可以使用类似蒸馏的方式进行无标签训练
# PPQ 模型的训练过程与 Pytorch 遵循相同的逻辑，你可以使用 Pytorch 中的技巧来获得更好的训练效果
# ------------------------------------------------------------
model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
graph = load_torch_model(model=model, sample=torch.zeros([1, 3, 224, 224]).cuda())

# ------------------------------------------------------------
# 创建自定义的 Quantizer, 并完成量化
# 你需要参考 esp_ppq.quantization.quantizer 文件夹里其他的 quantizer 定义来设计你的量化规则
# Quantizer 负责为所有算子初始化量化信息，你需要实现它的所有接口函数
# ------------------------------------------------------------
from esp_ppq.quantization.quantizer import BaseQuantizer


class MyQuantizer(BaseQuantizer):
    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        """
        函数 init_quantize_config 是 Quantizer 类的核心接口函数。PPQ 使用这个函数为所有算子初始化 TQC(Tensor Quantization Config)。
        通常而言，我们将根据算子的类型、算子的调度情况（算子上的 platform 属性）来决定如何量化给定的算子。

        你需要认识到一个事实，在 PPQ 中所有的逻辑都围绕着 TQC(Tensor Quantization Config) 这一控制结构体展开
            - Quantizer 负责初始化 TQC
            - Optim Passes 负责调整 TQC 的参数和状态
            - Exporter 负责导出 TQC

        init_quantize_config 函数只有一个参数，即需要被量化的算子，返回一个量化算子的量化控制结构体。

        Args:
            operation (Operation): 需要被量化的算子

        Returns:
            OperationQuantizationConfig: 返回的控制结构体
        """
        # ------------------------------------------------------------
        # 成员函数 create_default_quant_config 负责提供默认的 OQC (Operation Quantization Config)
        # 数据结构 OQC 只是 TQC 的一个简单集合，对于任何一个 onnx 算子而言，它都会具有 n 个输入变量和 m 个输出变量
        # 因此对应的 OQC 内将包含 n 的关于输入变量的 TQC, 以及 m 个关于输出变量的 TQC

        # 以卷积算子为例，它可以有 2 - 3 个输入变量，以及至多一个输出变量
        # 因此它的 OQC 内将含有 2 - 3 个输入 TQC，分别对应 输入量化、权重量化、bias 量化

        # OQC 一旦创建，则图结构不能发生改变，因此图的变换需要发生在量化之前
        # 因为图结构一旦变换，则可能导致算子的输入个数变化，从而导致 OQC 不匹配

        # 下面的代码将把算子上所有的 TQC 初始化为非对称 PER_TENSOR 量化
        # ------------------------------------------------------------
        OQC = self.create_default_quant_config(
            policy=QuantizationPolicy(
                QuantizationProperty.ASYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
            ),
            rounding=RoundingPolicy.ROUND_HALF_EVEN,
            op=operation,
            num_of_bits=8,
            exponent_bits=0,
            quant_max=255,
            quant_min=0,
            observer_algorithm='percentile',
        )

        # 对于卷积和矩阵乘算子而言，它的 bias 量化通常需要使用 scale = input scale * weight scale
        # 此时我们希望你将卷积的 bias 量化状态设置为 PASSIVE_INIT, 而后 PassiveParameterQuantizePass 将会正确处理 bias 的 scale 问题
        # 于此同时你还需要将 bias 的量化位宽调整到 32 位，并且相应的调整 quant min, quant max
        # 同时将 bias 的量化调整为 对称 PER_TENSOR 量化
        if operation.type in {'Conv', 'ConvTranspose', 'Gemm'}:
            if operation.num_of_input == 3:
                TQC = OQC.input_quantization_config[-1]
                TQC.state = QuantizationStates.PASSIVE_INIT
                TQC.num_of_bits = 32
                TQC.quant_max = 1 << 30
                TQC.quant_min = -1 << 30
                TQC.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
                )
                TQC.observer_algorithm = 'minmax'

        # 对于卷积和矩阵乘算子而言，它们的权重可以进行 PER_CHANNEL 量化，因此你可以将权重量化的策略调整为 PER_CHANNEL
        # 于此同时我们将权重量化策略调整为 对称 PER_CHANNEL
        # 相应地修改 quant min, quant max
        if operation.type in {'Conv', 'ConvTranspose', 'Gemm', 'MatMul'}:
            # 有些时候，矩阵乘法算子并没有权重，而是两路激活值相乘，因此我们需要单独判断这种情况
            if not operation.inputs[1].is_parameter:
                TQC = OQC.input_quantization_config[1]
                TQC.num_of_bits = 8
                TQC.quant_max = 127
                TQC.quant_min = -128
                TQC.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                )
                TQC.observer_algorithm = 'minmax'

        # 你也可以按算子的名字为算子创建不一样的量化信息
        # 下面的代码把名字叫 MyLayer 的层指定为 PER TENSOR 量化
        if operation.name == 'MyLayer' and operation.type in {'Conv', 'ConvTranspose', 'Gemm', 'MatMul'}:
            TQC = OQC.input_quantization_config[1]
            TQC.num_of_bits = 8
            TQC.quant_max = 127
            TQC.quant_min = -128
            TQC.policy = QuantizationPolicy(
                QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
            )
            TQC.observer_algorithm = 'minmax'

        # 如果算子类型不在支持的范围内，我们给出警告
        if operation.type not in self.quant_operation_types:
            print(f'Operation {operation.name}({operation.type}) is not supported with current Quantizer.')

        return OQC

    def quant_operation_types(self) -> set:
        """quant_operation_types 返回一个类型集合，该集合被用于调度和量化"""
        return {
            'Conv',
            'ConvTranspose',
            'Gemm',
            'Relu',
            'PRelu',
            'Clip',
            'Pad',
            'Resize',
            'MaxPool',
            'AveragePool',
            'GlobalMaxPool',
            'GlobalAveragePool',
            'Softmax',
            'Mul',
            'Add',
            'Max',
            'Sub',
            'Div',
            'Reshape',
            'LeakyRelu',
            'Concat',
            'Sigmoid',
            'Interp',
            'ReduceMean',
            'Transpose',
            'Slice',
            'Flatten',
            'HardSwish',
            'HardSigmoid',
            'MatMul',
        }

    def activation_fusion_types(self) -> set:
        """
        quant_operation_types 返回一个类型集合，该集合被用于图融合

        对于任何一个类型在 activation_fusion_types 中的算子而言
        PPQ QuantizeFusionPass 会探测 Conv + activation, Gemm + activation 等结构
        会取消 Conv, Gemm 的输出量化及 activation 的输入量化，从而模拟推理时可能发生的激活算子融合

        """
        return {'Clip', 'Relu', 'Sigmoid', 'Swish', 'SoftPlus', 'Gelu', 'LeakyRelu'}


quantizer = MyQuantizer(graph=graph)
dispatching_table = PFL.Dispatcher(graph=graph).dispatch(quantizer.quant_operation_types)

# 为算子初始化量化信息
for op in graph.operations.values():
    quantizer.quantize_operation(op_name=op.name, platform=dispatching_table[op.name])

executor = TorchExecutor(graph=graph)
executor.tracing_operation_meta(inputs=torch.zeros([1, 3, 224, 224]).cuda())

calib_dataloader = load_imagenet_from_directory(
    directory=CFG_TRAIN_DIR, batchsize=CFG_BATCHSIZE, shuffle=True, subset=1280, require_label=True, num_of_workers=8
)

training_dataloader = load_imagenet_from_directory(
    directory=CFG_TRAIN_DIR, batchsize=CFG_BATCHSIZE, shuffle=True, require_label=True, num_of_workers=8
)

eval_dataloader = load_imagenet_from_directory(
    directory=CFG_VALIDATION_DIR, batchsize=CFG_BATCHSIZE, shuffle=True, require_label=True, num_of_workers=8
)

with ENABLE_CUDA_KERNEL():
    # ------------------------------------------------------------
    # 创建优化管线，由于后续还要继续训练我们的模型，我们不能在此处调用
    # ParameterBakingPass()，一旦模型权重完成烘焙，则它们不能被进一步调整
    # ------------------------------------------------------------
    pipeline = PFL.Pipeline(
        [
            QuantizeSimplifyPass(),
            QuantizeFusionPass(activation_type=quantizer.activation_fusion_types),
            ParameterQuantizePass(),
            RuntimeCalibrationPass(method='kl'),
            PassiveParameterQuantizePass(),
            QuantAlignmentPass(),
            # LearnedStepSizePass(steps=500, block_size=5)
        ]
    )
    pipeline.optimize(
        calib_steps=8, collate_fn=lambda x: x[0].cuda(), graph=graph, dataloader=calib_dataloader, executor=executor
    )

    # ------------------------------------------------------------
    # 完成量化后，我们将开始进行 QAT 的模型训练，我们希望你能够注意到：
    #
    # 1. 不能从零开始使用 QAT 的方法完成训练，你应当先训练好浮点的模型，或者在一个预训练的模型基础上进行 QAT finetuning.
    # 2. 你必须完成标准量化流程
    # 3. PPQ Executor 长得很像 Pytorch Module，单机训练应该不会遇到太多困难，但它不支持多卡训练

    # 训练的代码我们封装进了一个叫做 ImageNetTrainer 的东西
    # 你可以打开它看到具体的训练逻辑
    # ------------------------------------------------------------
    trainer = ImageNetTrainer(graph=graph)

    best_acc = 0
    for epoch in range(20):
        trainer.epoch(training_dataloader)
        current_acc = trainer.eval(eval_dataloader)
        if current_acc > best_acc:
            trainer.save('Best.native')

    graph = load_native_graph(import_file='Best.native')
    PFL.Exporter(platform=CFG_PLATFORM).export(file_path='export.onnx', graph=graph, config_path='export.json')



# 这里我们展示两种不同的方法去生成 TensorRT Engine

# Plan B: PPQ 导出 engine

import os

import torchvision.transforms as transforms
from PIL import Image

from esp_ppq import *
from esp_ppq.api import *

ONNX_PATH = 'models/yolov5s6.onnx'  # 你的模型位置
OUTPUT_PATH = 'Output'  # 生成的量化模型的位置
CALIBRATION_PATH = 'imgs'  # 校准数据集
BATCHSIZE = 1
EXECUTING_DEVICE = 'cuda'
# create dataloader
imgs = []
trans = transforms.Compose(
    [
        transforms.Resize([640, 640]),  # [h,w]
        transforms.ToTensor(),
    ]
)
for file in os.listdir(path=CALIBRATION_PATH):
    path = os.path.join(CALIBRATION_PATH, file)
    img = Image.open(path).convert('RGB')
    img = trans(img)
    imgs.append(img)  # img is 0 - 1

from torch.utils.data import DataLoader

dataloader = DataLoader(dataset=imgs, batch_size=BATCHSIZE)

with ENABLE_CUDA_KERNEL():
    qir = quantize_onnx_model(
        platform=TargetPlatform.TRT_INT8,
        onnx_import_file=ONNX_PATH,
        calib_dataloader=dataloader,
        calib_steps=32,
        device=EXECUTING_DEVICE,
        input_shape=[BATCHSIZE, 3, 640, 640],
        collate_fn=lambda x: x.to(EXECUTING_DEVICE),
    )

    snr_report = graphwise_error_analyse(
        graph=qir, running_device=EXECUTING_DEVICE, dataloader=dataloader, collate_fn=lambda x: x.to(EXECUTING_DEVICE)
    )

    snr_report = layerwise_error_analyse(
        graph=qir, running_device=EXECUTING_DEVICE, dataloader=dataloader, collate_fn=lambda x: x.to(EXECUTING_DEVICE)
    )

    export_ppq_graph(
        qir,
        platform=TargetPlatform.TRT_INT8,
        graph_save_to=OUTPUT_PATH.join('/INT8.onnx'),
        config_save_to=OUTPUT_PATH.join('/INT8.json'),
    )

    from esp_ppq.utils.TensorRTUtil import Benchmark, Profiling, build_engine

    build_engine(
        onnx_file=OUTPUT_PATH.join('/INT8.onnx'),
        engine_file=OUTPUT_PATH.join('/INT8.engine'),
        int8=True,
        int8_scale_file=OUTPUT_PATH.join('/INT8.json'),
    )

    Benchmark(OUTPUT_PATH.join('/INT8.engine'))
    Profiling(OUTPUT_PATH.join('/INT8.engine'))



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class DimensionValue(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DimensionValue()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsDimensionValue(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # DimensionValue
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DimensionValue
    def DimType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # DimensionValue
    def DimValue(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # DimensionValue
    def DimParam(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def DimensionValueStart(builder):
    builder.StartObject(3)

def Start(builder):
    DimensionValueStart(builder)

def DimensionValueAddDimType(builder, dimType):
    builder.PrependInt8Slot(0, dimType, 0)

def AddDimType(builder, dimType):
    DimensionValueAddDimType(builder, dimType)

def DimensionValueAddDimValue(builder, dimValue):
    builder.PrependInt64Slot(1, dimValue, 0)

def AddDimValue(builder, dimValue):
    DimensionValueAddDimValue(builder, dimValue)

def DimensionValueAddDimParam(builder, dimParam):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(dimParam), 0)

def AddDimParam(builder, dimParam):
    DimensionValueAddDimParam(builder, dimParam)

def DimensionValueEnd(builder):
    return builder.EndObject()

def End(builder):
    return DimensionValueEnd(builder)


class DimensionValueT(object):

    # DimensionValueT
    def __init__(self):
        self.dimType = 0  # type: int
        self.dimValue = 0  # type: int
        self.dimParam = None  # type: str

    @classmethod
    def InitFromBuf(cls, buf, pos):
        dimensionValue = DimensionValue()
        dimensionValue.Init(buf, pos)
        return cls.InitFromObj(dimensionValue)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, dimensionValue):
        x = DimensionValueT()
        x._UnPack(dimensionValue)
        return x

    # DimensionValueT
    def _UnPack(self, dimensionValue):
        if dimensionValue is None:
            return
        self.dimType = dimensionValue.DimType()
        self.dimValue = dimensionValue.DimValue()
        self.dimParam = dimensionValue.DimParam()

    # DimensionValueT
    def Pack(self, builder):
        if self.dimParam is not None:
            dimParam = builder.CreateString(self.dimParam)
        DimensionValueStart(builder)
        DimensionValueAddDimType(builder, self.dimType)
        DimensionValueAddDimValue(builder, self.dimValue)
        if self.dimParam is not None:
            DimensionValueAddDimParam(builder, dimParam)
        dimensionValue = DimensionValueEnd(builder)
        return dimensionValue



MODEL = 'model.onnx'
INPUT_SHAPE = [1, 3, 224, 224]
SAMPLES = [torch.rand(size=[INPUT_SHAPE]) for _ in range(256)]  # rewirte this to use real data.

# -------------------------------------------------------------------
# 打开 trt_infer 看到具体细节，这个文件是 nvidia 的官方实例
# -------------------------------------------------------------------
from trt_infer import EngineBuilder

builder = EngineBuilder()
builder.create_network('model_fp32.onnx')
builder.create_engine(engine_path='model_fp32.engine', precision="fp16")

# -------------------------------------------------------------------
# 启动 tensorRT 进行推理，你先装一下 trt
# -------------------------------------------------------------------
import tensorrt as trt
import trt_infer

samples = [convert_any_to_numpy(sample) for sample in SAMPLES]
logger = trt.Logger(trt.Logger.INFO)
with open('model_fp32.engine', 'rb') as f, trt.Runtime(logger) as runtime:
    engine = runtime.deserialize_cuda_engine(f.read())

results = []
with engine.create_execution_context() as context:
    inputs, outputs, bindings, stream = trt_infer.allocate_buffers(context.engine)
    for sample in tqdm(samples, desc='TensorRT is running...'):
        inputs[0].host = convert_any_to_numpy(sample)
        [output] = trt_infer.do_inference(
            context, bindings=bindings, inputs=inputs, outputs=outputs, stream=stream, batch_size=1
        )
        results.append(convert_any_to_torch_tensor(output))



from typing import Union

import torch

from esp_ppq.api.setting import QuantizationSetting
from esp_ppq.core import (
    PASSIVE_OPERATIONS,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
)
from esp_ppq.IR import BaseGraph, Operation
from esp_ppq.quantization.optim import QuantizationOptimizationPipeline

from .base import BaseQuantizer


class FPGAQuantizer(BaseQuantizer):
    def __init__(
        self,
        graph: BaseGraph,
    ) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = -128
        self._quant_max = +127

    def build_quant_pipeline(self, setting: QuantizationSetting) -> QuantizationOptimizationPipeline:
        pipeline = super().build_quant_pipeline(setting)
        return pipeline

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='percentile',
        )

        if operation.type in {'Conv', 'ConvTranspose', 'Gemm'}:
            # set all parameters within Conv, ConvTranspose, Gemm to per-channel quant-config.
            assert operation.num_of_input > 0, 'Seems you got a Conv layer with no parameters.'

            # if operation has bias
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL
                    + QuantizationProperty.LINEAR
                    + QuantizationProperty.PER_TENSOR
                    + QuantizationProperty.POWER_OF_2
                )

                # Xilinx FPGA bias 并不是 32 位的！
                bias_config.num_of_bits = 30
                bias_config.quant_max = +int(pow(2, 29))
                bias_config.quant_min = -int(pow(2, 29))
                bias_config.state = QuantizationStates.PASSIVE_INIT
                base_quant_config.input_quantization_config[-1].observer_algorithm = 'Minmax'

        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            base_quant_config.is_active_quant_op = False

        return base_quant_config

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.FPGA_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quant_operation_types(self) -> set:
        return {
            'Conv',
            'ConvTranspose',
            'Gemm',
            'Relu',
            'PRelu',
            'Clip',
            'Pad',
            'Resize',
            'MaxPool',
            'AveragePool',
            'GlobalMaxPool',
            'GlobalAveragePool',
            'Mul',
            'Add',
            'Max',
            'Sub',
            'Div',
            'LeakyRelu',
            'Concat',
            'Sigmoid',
            'Slice',
        }

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL
            + QuantizationProperty.LINEAR
            + QuantizationProperty.PER_TENSOR
            + QuantizationProperty.POWER_OF_2
        )

    @property
    def rounding_policy(self):
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        """
        我不知道这个对不对, Xilinx FPGA 图融合做的并不少，
        如果你的网络中有特殊的激活函数，我建议你手动调整这个融合选项

        Returns:
            set: _description_
        """
        return {'Relu', 'Clip'}



from abc import abstractmethod
from typing import Dict, List, Set

from esp_ppq.core import TargetPlatform
from esp_ppq.IR import BaseGraph, Operation


class GraphDispatcher:
    """Graph Dispatcher cuts a graph into parts, each part of graph will
    dispatch to a specific platform for further execution and quantization.

    For the most part, all operations within graph can be partitioned into quantable operations,
        Shape-Or-Index related operations and remaining operations, all sub classes of GraphDispatcher will
        give an implementation of function "dispatch" to send all operations to their proper platform.

    ATTENTION: platform attribute will greatly affect quantizer's quantization logic, and the execution result.
        If operation is sent to a quantable platform, then its inputs and outputs will be quantized if necessary.
        if operation is classified as shape-or-index related operation, then its execution will be taken with cpu.
        if operation is sent to a fp32 platform, then its inputs and outputs shall never be quantized.
    """

    @abstractmethod
    def dispatch(self, quant_types: Set[str], **kwargs) -> Dict[str, TargetPlatform]:
        """Graph Dispatcher splits a graph into parts, each part of graph will
        be sent to a specific platform for further execution and quantization.
        """
        raise NotImplementedError('Impl this first.')


def value_tracing_pattern(from_where: Operation, to_where: Operation) -> bool:
    if to_where.type in {
        'Reshape',
        'Slice',
        'Gather',
        'Pad',
        'Resize',
        'Split',
        'TopK',
        'Tile',
        'Expand',
        'RoiAlign',
        'MMCVRoiAlign',
    }:
        # shape can go through above operations as a input, under this circumstance, their output should still be a tensor of shape.
        # however if shape was feed as a parameter for those operations, then their outputs are irrelevant with shape flow.
        return to_where.inputs[0].source_op == from_where
    if to_where.type == 'ScatterND':
        # ScatterND has 2 quant input.
        return to_where.inputs[0].source_op == from_where or to_where.inputs[-1].source_op == from_where

    if to_where.type in {'ConstantOfShape', 'Shape', 'NonMaxSuppression'}:
        # Inputs: (1)
        #   input : T
        #       1D tensor. The shape of the expected output tensor.
        #       If empty tensor is given, the output would be a scalar. All values must be >= 0.
        # see also: https://github.com/onnx/onnx/blob/master/docs/Operators.md#ConstantOfShape
        return False

    return True


def reverse_tracing_pattern(from_where: Operation, to_where: Operation) -> bool:
    if to_where.type in {'Shape'}:
        return False
    if to_where.type == 'TopK':
        return False
    if from_where.type in {
        'Reshape',
        'Slice',
        'Gather',
        'Pad',
        'Resize',
        'Split',
        'TopK',
        'Tile',
        'Expand',
        'RoiAlign',
        'MMCVRoiAlign',
    }:
        return to_where == from_where.inputs[0].source_op
    if from_where.type == 'ScatterND':
        return to_where == from_where.inputs[0].source_op or to_where == from_where.inputs[-1].source_op
    if from_where.type in {'NonMaxSuppression', 'Shape'}:  # 'ConstantOfShape'}:
        # remove constant of shape from here can speed up.
        return False
    return True


def SOI_receivers(graph: BaseGraph) -> Set[Operation]:
    receivers = set()
    for operation in graph.operations.values():
        for idx, plat in enumerate(operation.socket.in_plat):
            if plat in {TargetPlatform.SOI, TargetPlatform.FP32}:
                receivers.add(operation.inputs[idx].source_op)

    if None in receivers:
        receivers.remove(None)
    return receivers


def SOI_generators(graph: BaseGraph) -> Set[Operation]:
    _ret_collection = set()
    for operation in graph.operations.values():
        if operation.type in {'Shape', 'NonMaxSuppression', 'Constant', 'ConstantOfShape', 'TopK'}:
            _ret_collection.add(operation)
    return _ret_collection



from typing import Iterable

import torch
import torchvision

from esp_ppq import TargetPlatform, graphwise_error_analyse
from esp_ppq.api import quantize_torch_model
from esp_ppq.api.interface import ENABLE_CUDA_KERNEL
from esp_ppq.quantization.analyse.graphwise import statistical_analyse
from esp_ppq.quantization.analyse.layerwise import layerwise_error_analyse

# ------------------------------------------------------------
# 在 PPQ 中我们提供许多方法帮助你进行误差分析，误差分析是量化网络的第一步
# 这个脚本将以随机数据和 mobilenet v2 网络为例向你展示它们的使用方法
# ------------------------------------------------------------

BATCHSIZE = 32
INPUT_SHAPE = [BATCHSIZE, 3, 224, 224]
DEVICE = 'cuda'
PLATFORM = TargetPlatform.TRT_INT8


def load_calibration_dataset() -> Iterable:
    # ------------------------------------------------------------
    # 让我们从创建 calibration 数据开始做起， PPQ 需要你送入 32 ~ 1024 个样本数据作为校准数据集
    # 它们应该尽可能服从真实样本的分布，量化过程如同训练过程一样存在可能的过拟合问题
    # 你应当保证校准数据是经过正确预处理的、有代表性的数据，否则量化将会失败；校准数据不需要标签；数据集不能乱序
    # ------------------------------------------------------------
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]


CALIBRATION = load_calibration_dataset()


def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)


# ------------------------------------------------------------
# 我们使用 mobilenet v2 作为一个样例模型
# PPQ 将会使用 torch.onnx.export 函数 把 pytorch 的模型转换为 onnx 模型
# 对于复杂的 pytorch 模型而言，你或许需要自己完成 pytorch 模型到 onnx 的转换过程
# ------------------------------------------------------------
model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

# ------------------------------------------------------------
# 如果你使用 ENABLE_CUDA_KERNEL 方法
# PPQ 将会尝试编译自定义的高性能量化算子，这一过程需要编译环境的支持
# 如果你在编译过程中发生错误，你可以删除此处对于 ENABLE_CUDA_KERNEL 方法的调用
# 这将显著降低 PPQ 的运算速度；但即使你无法编译这些算子，你仍然可以使用 pytorch 的 gpu 算子完成量化
# ------------------------------------------------------------
with ENABLE_CUDA_KERNEL():
    quantized = quantize_torch_model(
        model=model,
        calib_dataloader=CALIBRATION,
        calib_steps=32,
        input_shape=INPUT_SHAPE,
        collate_fn=collate_fn,
        platform=PLATFORM,
        onnx_export_file='Output/onnx.model',
        device=DEVICE,
        verbose=0,
    )

    # ------------------------------------------------------------
    # graphwise_error_analyse 是最常用的分析方法，它分析网络中的量化误差情况，它的结果将直接打印在屏幕上
    # 对于 graphwise_error_analyse 而言，算子的误差直接衡量了量化网络与浮点网络之间的输出误差
    # 这一误差是累积的，意味着网络后面的算子总是会比网络前面的算子拥有更高的输出误差
    # 留意网络输出的误差情况，如果你想获得一个精度较高的量化网络，那么那些靠近输出的节点误差不应超过 10%
    # 该方法只衡量 Conv, Gemm 算子的误差情况，如果你对其余算子的误差情况感兴趣，需要手动修改方法逻辑
    # ------------------------------------------------------------
    reports = graphwise_error_analyse(
        graph=quantized, running_device=DEVICE, collate_fn=collate_fn, dataloader=CALIBRATION
    )

    # ------------------------------------------------------------
    # layerwise_error_analyse 是更为强大的分析方法，它分析算子的量化敏感性
    # 与 graphwise_error_analyse 不同，该方法分析的误差不是累计的
    # 该方法首先解除网络中所有算子的量化，而后单独地量化每一个 Conv, Gemm 算子
    # 以此来衡量量化单独一个算子对网络输出的影响情况，该方法常被用来决定网络调度与混合精度量化
    # 你可以将那些误差较大的层送往 TargetPlatform.FP32
    # ------------------------------------------------------------
    reports = layerwise_error_analyse(
        graph=quantized, running_device=DEVICE, collate_fn=collate_fn, dataloader=CALIBRATION
    )

    # ------------------------------------------------------------
    # statistical_analyse 是强有力的统计分析方法，该方法统计每一层的输入、输出以及参数的统计分布情况
    # 使用这一方法，你将更清晰地了解网络的量化情况，并能够有针对性地选择优化方案
    # 推荐在网络量化情况不佳时，使用 statistical_analyse 辅助你的分析
    # 该方法不打印任何数据，你需要手动将数据保存到硬盘并进行分析
    # ------------------------------------------------------------
    report = statistical_analyse(graph=quantized, running_device=DEVICE, collate_fn=collate_fn, dataloader=CALIBRATION)

    from pandas import DataFrame

    report = DataFrame(report)
    report.to_csv('1.csv')



import torch

from esp_ppq.core import QuantizationProperty, TensorQuantizationConfig

from .base import BaseQuantFunction
from .floating import PPQFloatingQuantFunction
from .linear import PPQDyamicLinearQuantFunction, PPQLinearQuant_toInt, PPQLinearQuantFunction


def PPQuantFunction(tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
    """
    ## PPQ 核心量化函数

    根据 config 中描述的策略，量化给定的 tensor.

    请注意 config.state 必须处在激活状态，该函数起作用。如果 config.state 处于
        INITIAL, FP32, PASSIVE_INIT 等未激活状态，该函数不进行任何处理，直接返回 tensor.

    ### 线性量化(QuantizationProperty.LINEAR):

        INT8 = Clip(Round((FP32 / scale)))

    ### 浮点量化(QuantizationProperty.FLOATING):

        FP8 = Clip(FP32_TO_FP8((FP32 / scale)))

    ### 动态线性量化(QuantizationProperty.DYNMAIC)

        scale = max(FP32) / 255

        INT8  = Clip(Round((FP32 / scale)))

    """
    if tensor is None:
        raise ValueError('Tensor is empty.')
    if config.policy.has_property(QuantizationProperty.LINEAR):
        if not config.policy.has_property(QuantizationProperty.DYNAMIC):
            return PPQLinearQuantFunction(tensor, config)
        else:
            return PPQDyamicLinearQuantFunction(tensor, config)

    if config.policy.has_property(QuantizationProperty.FLOATING):
        return PPQFloatingQuantFunction(tensor, config)

    raise ValueError(
        'Unexpected Quantization Property Found in PPQuantFunction. Do not konw how to quantize your config yet.'
    )


def PPQuantFunction_toInt(tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
    """
    ## PPQ 核心量化函数

    根据 config 中描述的策略，这个函数将会执行线性量化，动态量化

    但是结果直接出来是整数
    """

    if config.policy.has_property(QuantizationProperty.LINEAR):
        if not config.policy.has_property(QuantizationProperty.DYNAMIC):
            return PPQLinearQuant_toInt(tensor, config)

    raise ValueError(
        'Unexpected Quantization Property Found in PPQuantFunction_toInt. Do not konw how to quantize your config yet.'
    )


__all__ = [
    'PPQuantFunction',
    'PPQuantFunction_toInt',
    'PPQDyamicLinearQuantFunction',
    'PPQLinearQuantFunction',
    'PPQFloatingQuantFunction',
    'BaseQuantFunction',
    'PPQLinearQuant_toInt',
]



from typing import Iterable, List, Tuple

import torch

from esp_ppq.executor import BaseGraphExecutor
from esp_ppq.IR import BaseGraph, GraphFormatter, Operation, QuantableOperation, Variable
from esp_ppq.IR.morph import GraphDecomposer
from esp_ppq.log import NaiveLogger

from .base import QuantizationOptimizationPass

logger = NaiveLogger.get_logger('PPQ')


class NXPResizeModeChangePass(QuantizationOptimizationPass):
    """This optimization pass overwrite resize mode to 'nearest' for all resize
    operations."""

    def __init__(self) -> None:
        super().__init__(name='NXP Resize Operation Transformation')

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        for op in graph.operations.values():
            if op.type == 'Resize':
                op.attributes['mode'] = 'nearest'
                op.attributes['coordinate_transformation_mode'] = 'half_pixel'


class NCNNFormatGemmPass(QuantizationOptimizationPass):
    def __init__(self, name: str = 'ncnn Format Gemm Pass') -> None:
        super().__init__(name)

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        for op in graph.operations.values():
            if op.type == 'Gemm':
                if op.attributes.get('transB', 0) == 0:
                    op.attributes['transB'] = 1
                    weight = op.parameters[0].value
                    assert isinstance(weight, torch.Tensor)
                    op.parameters[0].value = weight.transpose(1, 0).contiguous()
                if op.attributes.get('alpha', 1.0) != 1.0:
                    op.parameters[0].value = op.parameters[0].value * op.attributes.get('alpha', 1.0)
                    op.attributes['alpha'] = 1.0
                if op.attributes.get('beta', 1.0) != 1.0:
                    if op.num_of_input > 2:
                        op.parameters[1].value = op.parameters[1].value * op.attributes.get('beta', 1.0)
                    op.attributes['beta'] = 1.0


class HorizontalLayerSplitPass(QuantizationOptimizationPass):
    """
    Horizontal Layer Split Pass(算子分裂过程)

    Split convolution layers or GEMM layers for better performance.

    Formula:

            Y = W * X + b

            where W can be divided into W_1 + W_2

            Y = (W_1 * X + b) + (W_2 * X)

    By splitting W like this, we are able to represent W more accurately.
    In the case where one channel has weights in the range [-32, 32] and another channel has weights in the range [-0.5, 0.5].
    the large channel will be divided so the range will come to [-16, 16], which leads us to use scale = 0.125 for representing
    the weight tensor rather than 0.25.

    The Estimation of Quantization Error is shown as a quadratic function of scale:

            E(Quantization Error) = scale ^ 2 / 12

    This Formula is proved by Bernard Widrow, according to the formula, a scale = 0.125 will decrease the quantization error by 75%.

    All the value larger than value_threshold will be divided into 2 part via this function, thus the layer itself will be
    splitted, an new Add operation are going to be created.

    ### Parameters:
        self.interested_layers = interested_layers
        self.value_threshold   = value_threshold
        self.method            = str(method).lower()
        self.verbose           = verbose

    # interested_layers(List[str])

            Only layer that listed in interested_layers will be processed by this pass.

            If interested_layers is None or empty list, NO layer will be processed.

    # value_threshold(float)

            This pass split value only when value is larger than value_threshold

            If there is no value large enough to be processed, corresponding layer will be skipped.

    # method(str)

            Splitting method, 'balance' or 'random'

            With balance method, W_1 and W_2 will be evenly divided.

            With random method, W_1 and W_2 will be randomly divided.

    ### Warning:

    Creating new operation in your network probably slows down the execution.

    Thus horizontal splitting is somehow a trade-off between speed and accuracy.

    ### Usage

    You can create this optimization manually:

        from esp_ppq import HorizontalLayerSplitPass

        optim = HorizontalLayerSplitPass()
    """

    def __init__(
        self,
        interested_layers: List[str] = None,
        value_threshold: float = 1,
        method: str = 'balance',
        verbose: bool = True,
    ) -> None:
        super().__init__('Layer Split Pass(Lateral)')
        self.interested_layers = interested_layers
        self.value_threshold = value_threshold
        self.method = str(method).lower()
        self.verbose = verbose

        if self.interested_layers is None or len(self.interested_layers) == 0:
            raise ValueError(
                'Layer Split Pass(Lateral) Requires a list of spliting layers, '
                'while parameter interested_layers is empty.'
            )

        if self.method not in {'balance', 'random'}:
            raise ValueError(f'Split method must be balance or random. While {self.method} is given.')

    def h_split(self, op: Operation) -> Tuple[torch.Tensor, torch.Tensor, int]:
        # split weight
        value = op.inputs[1].value
        mask = value.abs() > self.value_threshold
        processed_values = mask.sum().item()

        s_value = value
        if self.method == 'balance':
            s_value = (value / 2) * mask
        elif self.method == 'random':
            s_value = (value * torch.rand_like(value)) * mask
        else:
            raise Exception('Oops, seems we got some troubles here.')
        r_value = value - s_value

        # print
        if self.verbose:
            print('')
            print(f'# Layer {op.name} has been splited, {processed_values}/{value.numel()} value(s) was processed.')
        return r_value, s_value, processed_values

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        with torch.no_grad():
            for name in self.interested_layers:
                # op check
                if name not in graph.operations:
                    raise KeyError(f'Operation {name} is not in current graph.')
                op1 = graph.operations[name]
                if op1.type not in {'Gemm', 'MatMul', 'Conv', 'ConvTranspose'}:
                    raise TypeError(f'Operation {op1.name} can not be splited, op type is invalid({op1.type})')
                if not op1.inputs[1].is_parameter:
                    raise ValueError(f'Operation {op1.name} can not be splited, input 1 is not parameter.')
                if isinstance(op1, QuantableOperation):
                    raise TypeError(
                        'Can not split a quantized operation, '
                        'Layer Split Pass should only be invoked as a pre-quant optimziation.'
                    )

                r_value, s_value, processed_values = self.h_split(op1)

                if processed_values > 0:
                    # clone current operation
                    op2 = graph.create_operation(
                        op_type=op1.type, attributes=op1.attributes.copy(), platform=op1.platform
                    )
                    input_var, output_var = op1.inputs[0], op1.outputs[0]
                    graph.create_link_with_op(variable=op1.inputs[0], A=input_var.source_op, B=op2)

                    # create weight for cloned operation.
                    graph.create_variable(value=op1.inputs[1].value.clone(), is_parameter=True, dest_ops=[op2])

                    # set splited value
                    op1.inputs[1].value.copy_(r_value)
                    op2.inputs[1].value.copy_(s_value)

                    op1.outputs.clear()
                    adder = graph.create_operation(op_type='Add', platform=op1.platform, outputs=[output_var])
                    output_var.source_op = adder

                    graph.create_link_with_op(A=op1, B=adder)
                    graph.create_link_with_op(A=op2, B=adder)


class MetaxGemmSplitPass(QuantizationOptimizationPass):
    """Metax 不支持 Gemm 的量化，这个 pass 将 Gemm 拆分成.

    --- MatMul -----|
                    + --- Add ---
        bias   -----|
    """

    def __init__(self, name: str = 'Metax Gemm Split Pass') -> None:
        super().__init__(name)

    # Implementation of Gemm Split will move to IR.morph soon.
    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        morpher = GraphDecomposer(graph)
        morpher.decompose_gemm()


class GRUSplitPass(QuantizationOptimizationPass):
    """执行 GRU 算子分解，这个 Pass 将 GRU 算子分解为单步执行的形式.

    请注意，对于 ONNX GRU 算子而言, 它有两个输出, 一个是完整的hidden vector, 另一个是单步的 last state 这个
    pass 是针对单步执行而设计的，它将直接删除 hidden vector 之后的所有输出
    """

    def __init__(self, name: str = 'Metax Gemm Split Pass') -> None:
        super().__init__(name)

    def delete_hidden_vec(self, graph: BaseGraph, hidden_vec: Variable):
        processor = GraphFormatter(graph)
        processor.truncate_on_var(var=hidden_vec, mark_as_output=False)

    # Implementation of Gemm Split will move to IR.morph soon.
    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        interested_ops = []
        for operation in graph.operations.values():
            if operation.type == 'GRU':
                interested_ops.append(operation)

        for op in interested_ops:
            assert isinstance(op, Operation)
            # fetch all related variables
            rnn_x, rnn_w, rnn_r, rnn_b, _, rnn_h = op.inputs
            hidden_size = op.attributes['hidden_size']

            # Take a further look at
            # https://github.com/onnx/onnx/blob/main/docs/Operators.md#GRU
            Wz = rnn_w.value[0, hidden_size * 0 : hidden_size * 1]
            Wr = rnn_w.value[0, hidden_size * 1 : hidden_size * 2]
            Wh = rnn_w.value[0, hidden_size * 2 : hidden_size * 3]

            Rz = rnn_r.value[0, hidden_size * 0 : hidden_size * 1]
            Rr = rnn_r.value[0, hidden_size * 1 : hidden_size * 2]
            Rh = rnn_r.value[0, hidden_size * 2 : hidden_size * 3]

            Wbz = rnn_b.value[0, hidden_size * 0 : hidden_size * 1]
            Wbr = rnn_b.value[0, hidden_size * 1 : hidden_size * 2]
            Wbh = rnn_b.value[0, hidden_size * 2 : hidden_size * 3]

            Rbz = rnn_b.value[0, hidden_size * 3 : hidden_size * 4]
            Rbr = rnn_b.value[0, hidden_size * 4 : hidden_size * 5]
            Rbh = rnn_b.value[0, hidden_size * 5 : hidden_size * 6]

            # create operations
            op1 = graph.create_operation(op_type='Gemm', attributes={'transB': 1})
            op2 = graph.create_operation(op_type='Gemm', attributes={'transB': 1})
            op3 = graph.create_operation(op_type='Add')
            op4 = graph.create_operation(op_type='Sigmoid')
            op5 = graph.create_operation(op_type='Slice')
            op6 = graph.create_operation(op_type='Slice')
            op7 = graph.create_operation(op_type='Gemm', attributes={'transB': 1})
            op8 = graph.create_operation(op_type='Gemm', attributes={'transB': 1})
            op9 = graph.create_operation(op_type='Mul')
            op10 = graph.create_operation(op_type='Mul')
            op11 = graph.create_operation(op_type='Sub')
            op12 = graph.create_operation(op_type='Add')
            op13 = graph.create_operation(op_type='Mul')
            op14 = graph.create_operation(op_type='Tanh')
            op15 = graph.create_operation(op_type='Add')

            # create parameter variables
            # 为了加速运算，我们将Wz, Wr合并成Wzr, Rzh等同理
            # 参考 https://github.com/onnx/onnx/blob/main/docs/Operators.md#GRU
            Wzr_var = graph.create_variable(value=torch.cat([Wz, Wr]), is_parameter=True)
            Rzr_var = graph.create_variable(value=torch.cat([Rz, Rr]), is_parameter=True)
            Wbzr_var = graph.create_variable(value=torch.cat([Wbz, Wbr]), is_parameter=True)
            Rbzr_var = graph.create_variable(value=torch.cat([Rbz, Rbr]), is_parameter=True)

            Wh_var = graph.create_variable(value=Wh, is_parameter=True)
            Rh_var = graph.create_variable(value=Rh, is_parameter=True)
            Wbh_var = graph.create_variable(value=Wbh, is_parameter=True)
            Rbh_var = graph.create_variable(value=Rbh, is_parameter=True)

            constant_of_sub = graph.create_variable(value=torch.tensor(1.0).to(Wz.device), is_parameter=True)

            # link variables
            graph.create_link_with_op(variable=constant_of_sub, A=None, B=op11)
            graph.create_link_with_op(A=op1, B=op3)
            graph.create_link_with_op(A=op2, B=op3)
            graph.create_link_with_op(A=op3, B=op4)

            var = graph.create_variable()
            graph.create_link_with_op(variable=var, A=op4, B=op5)
            graph.create_link_with_op(variable=var, A=op4, B=op6)

            var = graph.create_variable()
            graph.create_link_with_op(variable=var, A=op5, B=op11)
            graph.create_link_with_op(variable=var, A=op5, B=op10)

            graph.create_link_with_op(A=op6, B=op9)
            graph.create_link_with_op(A=op7, B=op9)
            graph.create_link_with_op(A=op8, B=op12)
            graph.create_link_with_op(A=op9, B=op12)
            graph.create_link_with_op(A=op10, B=op15)
            graph.create_link_with_op(A=op11, B=op13)
            graph.create_link_with_op(A=op12, B=op14)
            graph.create_link_with_op(A=op13, B=op15)
            graph.create_link_with_op(A=op14, B=op13)

            # mark h as graph input, link h to op2, op10 and op7
            rnn_h.source_op.outputs.remove(rnn_h)
            rnn_h.source_op = None
            rnn_h.dest_ops.remove(op)
            graph.mark_variable_as_graph_input(rnn_h)
            graph.create_link_with_op(variable=rnn_h, A=None, B=op2)
            graph.create_link_with_op(variable=rnn_h, A=None, B=op7)
            graph.create_link_with_op(variable=rnn_h, A=None, B=op10)

            # link x to op1 and op8
            rnn_x.dest_ops.remove(op)
            graph.create_link_with_op(variable=rnn_x, A=rnn_x.source_op, B=op1)
            graph.create_link_with_op(variable=rnn_x, A=rnn_x.source_op, B=op8)

            # create parameters
            graph.create_link_with_op(variable=Wzr_var, A=None, B=op1)
            graph.create_link_with_op(variable=Rzr_var, A=None, B=op2)
            graph.create_link_with_op(variable=Wh_var, A=None, B=op8)
            graph.create_link_with_op(variable=Rh_var, A=None, B=op7)
            graph.create_link_with_op(variable=Wbzr_var, A=None, B=op1)
            graph.create_link_with_op(variable=Rbzr_var, A=None, B=op2)
            graph.create_link_with_op(variable=Wbh_var, A=None, B=op8)
            graph.create_link_with_op(variable=Rbh_var, A=None, B=op7)

            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([0]), is_parameter=True), A=None, B=op5
            )
            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([hidden_size]), is_parameter=True), A=None, B=op5
            )
            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([1]), is_parameter=True), A=None, B=op5
            )
            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([1]), is_parameter=True), A=None, B=op5
            )

            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([hidden_size]), is_parameter=True), A=None, B=op6
            )
            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([2 * hidden_size]), is_parameter=True), A=None, B=op6
            )
            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([1]), is_parameter=True), A=None, B=op6
            )
            graph.create_link_with_op(
                variable=graph.create_variable(value=torch.tensor([1]), is_parameter=True), A=None, B=op6
            )

            hidden_vec, last_state = op.outputs
            last_state.source_op = op15
            op15.outputs.append(last_state)

            op.inputs.clear()
            op.outputs.clear()
            graph.remove_operation(op)
            self.delete_hidden_vec(graph, hidden_vec)


class ConvTransposeDecompositionPass(QuantizationOptimizationPass):
    """Decompose ConvTranspose operations into InsertZeros + Conv.

    This pass replaces ConvTranspose operations with a combination of
    InsertZeros operation (for stride) followed by a standard Conv operation
    with rotated and transposed kernel.

    This decomposition is useful for platforms that don't support ConvTranspose
    natively but support standard Conv operations.
    """

    def __init__(self, name: str = 'ConvTranspose Decomposition Pass') -> None:
        super().__init__(name)

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        from esp_ppq.IR.morph import GraphDecomposer

        decomposer = GraphDecomposer(graph)
        decomposer.decompose_convtranspose()



import os

import numpy as np
import torch
import torchvision

import esp_ppq.lib as PFL
from esp_ppq import TargetPlatform, TorchExecutor, graphwise_error_analyse
from esp_ppq.api import ENABLE_CUDA_KERNEL, export_ppq_graph, load_torch_model
from esp_ppq.core import convert_any_to_numpy
from esp_ppq.quantization.optim import *

calibration_dataloader = []
for file in os.listdir('imagenet'):
    path = os.path.join('imagenet', file)
    arr = np.fromfile(path, dtype=np.dtype('float32')).reshape([1, 3, 224, 224])
    calibration_dataloader.append(torch.tensor(arr))

with ENABLE_CUDA_KERNEL():
    model = torchvision.models.mnasnet1_0(pretrained=True).cuda()
    graph = load_torch_model(model=model, sample=torch.zeros(size=[1, 3, 224, 224]).cuda())
    # ------------------------------------------------------------
    # 我们首先进行标准的量化流程，为所有算子初始化量化信息，并进行 Calibration
    # ------------------------------------------------------------
    quantizer = PFL.Quantizer(platform=TargetPlatform.TRT_INT8, graph=graph)  # 取得 TRT_INT8 所对应的量化器
    dispatching = PFL.Dispatcher(graph=graph).dispatch(  # 生成调度表
        quant_types=quantizer.quant_operation_types
    )

    # 为算子初始化量化信息
    for op in graph.operations.values():
        quantizer.quantize_operation(op_name=op.name, platform=dispatching[op.name])

    # 初始化执行器
    collate_fn = lambda x: x.to('cuda')
    executor = TorchExecutor(graph=graph, device='cuda')
    executor.tracing_operation_meta(inputs=torch.zeros(size=[1, 3, 224, 224]).cuda())
    executor.load_graph(graph=graph)

    # ------------------------------------------------------------
    # 创建优化管线，由于后续还要继续训练我们的模型，我们不能在此处调用
    # ParameterBakingPass()，一旦模型权重完成烘焙，则它们不能被进一步调整
    # ------------------------------------------------------------
    pipeline = PFL.Pipeline(
        [
            QuantizeSimplifyPass(),
            QuantizeFusionPass(activation_type=quantizer.activation_fusion_types),
            ParameterQuantizePass(),
            RuntimeCalibrationPass(),
            PassiveParameterQuantizePass(),
            QuantAlignmentPass(force_overlap=True),
        ]
    )

with ENABLE_CUDA_KERNEL():
    # 调用管线完成量化
    pipeline.optimize(
        graph=graph,
        dataloader=calibration_dataloader,
        verbose=True,
        calib_steps=32,
        collate_fn=collate_fn,
        executor=executor,
    )

    graphwise_error_analyse(
        graph=graph, running_device='cuda', dataloader=calibration_dataloader, collate_fn=lambda x: x.cuda()
    )

    export_ppq_graph(
        graph=graph,
        platform=TargetPlatform.TRT_INT8,
        graph_save_to='Output/quantized.onnx',
        config_save_to='Output/quantized.json',
    )

    results, executor = [], TorchExecutor(graph=graph)
    for idx, data in enumerate(calibration_dataloader):
        arr = convert_any_to_numpy(executor(data.cuda())[0])
        arr.tofile(f'Output/Result/{idx}.bin')

    from esp_ppq.utils.TensorRTUtil import build_engine

    build_engine(
        onnx_file='Output/quantized.onnx',
        int8_scale_file='Output/quantized.json',
        engine_file='Output/INT8.engine',
        int8=True,
        fp16=True,
    )
    build_engine(onnx_file='Output/quantized.onnx', engine_file='Output/FP16.engine', int8=False, fp16=True)
    build_engine(onnx_file='Output/quantized.onnx', engine_file='Output/FP32.engine', int8=False, fp16=False)
